# -*- coding: utf-8 -*-
"""Dummy

This is designed to fail
"""

from km3pipe import Module


class Dummy(Module):
    """Dummy base class.

    """
    def __init__(self, **kwargs):
        super(Dummy, self).__init__(**kwargs)
        self.hit_sel = self.get('hit_sel') or 'Hits'
        self.key_out = self.get('key_out') or 'PrimFit'

    def process(self, blob):
        print(self.fit())
        return blob

    def fit(self):
        return 42


class SubDummy(Dummy):
    """Primitive Sub-Subclass

    """
    def __init__(self, **kwargs):
        super(SubDummy, self).__init__(**kwargs)
        self.hit_sel = self.get('hit_sel') or 'Hits'
        self.key_out = self.get('key_out') or 'PrimFit'

    def fit(self):
        return 4200



__author__ = 'tamasgal'

from km3pipe import Module, Pipeline
from km3pipe.io import EvtPump


class PrintBlob(Module):
    def process(self, blob):
        print(blob.keys())
        return blob


pipeline = Pipeline()
pipeline.attach(EvtPump, 'evtpump', filename='files/example_numuNC.evt')
pipeline.attach(PrintBlob, 'printer')
pipeline.drain(1)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# vim: ts=4 sw=4 et
"""
=================
Live Event Dumper
=================

Recieves triggered events from the detector and dumps them to a file.

"""

from km3pipe import Pipeline, Module
from km3pipe.io.ch import CHPump
from km3pipe.io.daq import DAQProcessor

__author__ = 'Tamas Gal <tgal@km3net.de>'


class Dumper(Module):
    def configure(self):
        self.counter = 0

    def process(self, blob):
        tag = str(blob['CHPrefix'].tag)
        data = blob['CHData']
        self.dump(data, tag)
        print(blob["Hits"])
        return blob

    def dump(self, data, tag):
        with open('{0}-{1:06}.dat'.format(tag, self.counter), 'bw') as f:
            self.counter += 1
            f.write(data)


pipe = Pipeline()
pipe.attach(
    CHPump,
    host='xxx.xxx.xxx.xxx',
    port=5553,
    tags='IO_EVT',
    timeout=60 * 60 * 24,
    max_queue=42
)
pipe.attach(DAQProcessor)
pipe.attach(Dumper)
pipe.drain()



# Filename: test_time.py
# pylint: disable=locally-disabled,C0111,R0904,C0103

import sqlite3
import tempfile

import km3pipe as kp
from km3pipe.dataclasses import Table
from km3modules.common import (
    Siphon, Delete, Keep, Dump, StatusBar, TickTock, MemoryObserver,
    BlobIndexer, LocalDBService
)
from km3pipe.testing import TestCase, MagicMock
from km3pipe.tools import istype

__author__ = "Tamas Gal"
__copyright__ = "Copyright 2016, Tamas Gal and the KM3NeT collaboration."
__credits__ = []
__license__ = "MIT"
__maintainer__ = "Tamas Gal"
__email__ = "tgal@km3net.de"
__status__ = "Development"


class InfinitePump(kp.Pump):
    """A pump which just infinetly spits out indexed blobs"""
    def configure(self):
        self.i = 0

    def process(self, blob):
        self.i += 1
        blob['i'] = self.i
        return blob


class TestKeep(TestCase):
    def test_keep_a_single_key(self):
        class APump(kp.Pump):
            def process(self, blob):
                blob['a'] = 'a'
                blob['b'] = 'b'
                blob['c'] = 'c'
                blob['d'] = 'd'
                return blob

        class Observer(kp.Module):
            def process(self, blob):
                assert 'a' not in blob
                assert 'b' not in blob
                assert 'c' not in blob
                assert 'd' == blob['d']
                return blob

        pipe = kp.Pipeline()
        pipe.attach(APump)
        pipe.attach(Keep, keys='d')
        pipe.attach(Observer)
        pipe.drain(5)

    def test_keep_multiple_keys(self):
        class APump(kp.Pump):
            def process(self, blob):
                blob['a'] = 'a'
                blob['b'] = 'b'
                blob['c'] = 'c'
                blob['d'] = 'd'
                return blob

        class Observer(kp.Module):
            def process(self, blob):
                assert 'a' not in blob
                assert 'b' == blob['b']
                assert 'c' not in blob
                assert 'd' == blob['d']
                return blob

        pipe = kp.Pipeline()
        pipe.attach(APump)
        pipe.attach(Keep, keys=['b', 'd'])
        pipe.attach(Observer)
        pipe.drain(5)

    def test_hdf5_keep_group_wo_subgroup(self):
        class APump(kp.Pump):
            def process(self, blob):
                blob['A'] = kp.Table({
                    'foo': [1, 2, 3],
                    'bar': [4, 5, 6]
                },
                                     h5loc='/foobar')
                blob['B'] = kp.Table({
                    'a': [1.1, 2.1, 3.1],
                    'b': [4.2, 5.2, 6.2]
                },
                                     h5loc='/ab')
                return blob

        class Observer(kp.Module):
            def process(self, blob):
                assert 'A' in blob.keys()
                assert '/foobar' == blob['A'].h5loc
                assert not 'B' in blob.keys()
                return blob

        pipe = kp.Pipeline()
        pipe.attach(APump)
        pipe.attach(Keep, h5locs=['/foobar'])
        pipe.attach(Observer)
        pipe.drain(5)

    def test_hdf5_keep_group_w_subgroup(self):
        class APump(kp.Pump):
            def process(self, blob):
                blob['A'] = kp.Table({
                    'foo': [1, 2, 3],
                    'bar': [4, 5, 6]
                },
                                     h5loc='/foobar')
                blob['B'] = kp.Table({
                    'a': [1.1, 2.1, 3.1],
                    'b': [4.2, 5.2, 6.2]
                },
                                     h5loc='/ab')
                return blob

        class Observer(kp.Module):
            def process(self, blob):
                assert 'A' in blob.keys()
                assert '/foobar' == blob['A'].h5loc
                assert not 'B' in blob.keys()
                return blob

        pipe = kp.Pipeline()
        pipe.attach(APump)
        pipe.attach(Keep, h5locs=['/foobar'])
        pipe.attach(Observer)
        pipe.drain(5)

    def test_key_hdf5_group_individual(self):
        class APump(kp.Pump):
            def process(self, blob):
                blob['A'] = kp.Table({
                    'foo': [1, 2, 3],
                    'bar': [4, 5, 6]
                },
                                     h5loc='/foobar')
                blob['B'] = kp.Table({
                    'a': [1.1, 2.1, 3.1],
                    'b': [4.2, 5.2, 6.2]
                },
                                     h5loc='/ab')
                return blob

        class Observer(kp.Module):
            def process(self, blob):
                assert 'A' in blob.keys()
                assert '/foobar' == blob['A'].h5loc
                assert 'B' in blob.keys()
                assert '/ab' == blob['B'].h5loc
                return blob

        pipe = kp.Pipeline()
        pipe.attach(APump)
        pipe.attach(Keep, keys=['B'], h5locs=['/foobar'])
        pipe.attach(Observer)
        pipe.drain(5)

    def test_key_hdf5_group_parallel(self):
        class APump(kp.Pump):
            def process(self, blob):
                blob['A'] = kp.Table({
                    'foo': [1, 2, 3],
                    'bar': [4, 5, 6]
                },
                                     h5loc='/foobar')
                blob['B'] = kp.Table({
                    'a': [1.1, 2.1, 3.1],
                    'b': [4.2, 5.2, 6.2]
                },
                                     h5loc='/ab')
                return blob

        class Observer(kp.Module):
            def process(self, blob):
                assert 'A' in blob.keys()
                assert '/foobar' == blob['A'].h5loc
                assert not 'B' in blob.keys()
                return blob

        pipe = kp.Pipeline()
        pipe.attach(APump)
        pipe.attach(Keep, keys=['A'], h5locs=['/foobar'])
        pipe.attach(Observer)
        pipe.drain(5)

    def test_major_hdf5_group(self):
        class APump(kp.Pump):
            def process(self, blob):
                blob['A'] = kp.Table({
                    'foo': [1, 2, 3],
                    'bar': [4, 5, 6]
                },
                                     h5loc='/foobar/a')
                blob['B'] = kp.Table({
                    'a': [1.1, 2.1, 3.1],
                    'b': [4.2, 5.2, 6.2]
                },
                                     h5loc='/foobar/b')
                return blob

        class Observer(kp.Module):
            def process(self, blob):
                assert 'A' in blob.keys()
                assert '/foobar/a' == blob['A'].h5loc
                assert 'B' in blob.keys()
                assert '/foobar/b' == blob['B'].h5loc
                return blob

        pipe = kp.Pipeline()
        pipe.attach(APump)
        pipe.attach(Keep, h5locs=['/foobar'])
        pipe.attach(Observer)
        pipe.drain(5)

    def test_major_hdf5_group_nested(self):
        class APump(kp.Pump):
            def process(self, blob):
                blob['A'] = kp.Table({'a': 0}, h5loc='/foo/bar/a')
                blob['B'] = kp.Table({'b': 1}, h5loc='/foo/bar/baz/b')
                blob['C'] = kp.Table({'c': 2}, h5loc='/foo/bar/baz/fjord/c')
                return blob

        class Observer(kp.Module):
            def process(self, blob):
                assert 'A' not in blob
                assert 'B' in blob
                assert 'C' in blob
                return blob

        pipe = kp.Pipeline()
        pipe.attach(APump)
        pipe.attach(Keep, h5locs=['/foo/bar/baz'])
        pipe.attach(Observer)
        pipe.drain(5)


class TestDelete(TestCase):
    def test_delete_a_single_key(self):
        class APump(kp.Pump):
            def process(self, blob):
                blob['a'] = 'a'
                blob['b'] = 'b'
                blob['c'] = 'c'
                return blob

        class Observer(kp.Module):
            def process(self, blob):
                assert 'a' == blob['a']
                assert 'b' not in blob
                assert 'c' == blob['c']
                return blob

        pipe = kp.Pipeline()
        pipe.attach(APump)
        pipe.attach(Delete, key='b')
        pipe.attach(Observer)
        pipe.drain(5)

    def test_delete_multiple_keys(self):
        class APump(kp.Pump):
            def process(self, blob):
                blob['a'] = 'a'
                blob['b'] = 'b'
                blob['c'] = 'c'
                return blob

        class Observer(kp.Module):
            def process(self, blob):
                assert 'a' not in blob
                assert 'b' not in blob
                assert 'c' == blob['c']
                return blob

        pipe = kp.Pipeline()
        pipe.attach(APump)
        pipe.attach(Delete, keys=['a', 'b'])
        pipe.attach(Observer)
        pipe.drain(5)


class TestSiphon(TestCase):
    def test_siphon(self):
        class Observer(kp.Module):
            def configure(self):
                self.mock = MagicMock()

            def process(self, blob):
                self.mock()
                return blob

            def finish(self):
                assert self.mock.call_count == 7

        pipe = kp.Pipeline()
        pipe.attach(InfinitePump)
        pipe.attach(Siphon, volume=10)
        pipe.attach(Observer)
        pipe.drain(17)

    def test_siphon_with_flush(self):
        class Observer(kp.Module):
            def configure(self):
                self.mock = MagicMock()

            def process(self, blob):
                self.mock()
                return blob

            def finish(self):
                assert self.mock.call_count == 1

        pipe = kp.Pipeline()
        pipe.attach(InfinitePump)
        pipe.attach(Siphon, volume=10, flush=True)
        pipe.attach(Observer)
        pipe.drain(21)

    def test_siphon_with_flush_2(self):
        class Observer(kp.Module):
            def configure(self):
                self.mock = MagicMock()

            def process(self, blob):
                self.mock()
                return blob

            def finish(self):
                assert self.mock.call_count == 2

        pipe = kp.Pipeline()
        pipe.attach(InfinitePump)
        pipe.attach(Siphon, volume=10, flush=True)
        pipe.attach(Observer)
        pipe.drain(22)


class TestDump(TestCase):
    def test_dump(self):
        def add_something(blob):
            blob['a'] = 1
            return blob

        pipe = kp.Pipeline()
        pipe.attach(InfinitePump)
        pipe.attach(add_something)
        pipe.attach(Dump)
        pipe.drain(3)

    def test_dump_a_key(self):
        def add_something(blob):
            blob['a'] = 1
            return blob

        pipe = kp.Pipeline()
        pipe.attach(InfinitePump)
        pipe.attach(add_something)
        pipe.attach(Dump, key='a')
        pipe.drain(3)

    def test_dump_multiple_keys(self):
        def add_something(blob):
            blob['a'] = 1
            blob['b'] = 2
            return blob

        pipe = kp.Pipeline()
        pipe.attach(InfinitePump)
        pipe.attach(add_something)
        pipe.attach(Dump, keys=['a', 'b'])
        pipe.drain(3)

    def test_dump_full(self):
        def add_something(blob):
            blob['a'] = 1
            blob['b'] = 2
            return blob

        pipe = kp.Pipeline()
        pipe.attach(InfinitePump)
        pipe.attach(add_something)
        pipe.attach(Dump, full=True)
        pipe.drain(3)


class TestStatusbar(TestCase):
    def test_statusbar(self):
        pipe = kp.Pipeline()
        pipe.attach(InfinitePump)
        pipe.attach(StatusBar, every=2)
        pipe.drain(5)


class TestTickTock(TestCase):
    def test_ticktock(self):
        pipe = kp.Pipeline()
        pipe.attach(InfinitePump)
        pipe.attach(TickTock)
        pipe.drain(5)


class TestMemoryObserver(TestCase):
    def test_memory_observer(self):
        pipe = kp.Pipeline()
        pipe.attach(InfinitePump)
        pipe.attach(MemoryObserver)
        pipe.drain(5)


class TestBlobIndexer(TestCase):
    def test_blob_indexer(self):
        class Observer(kp.Module):
            def configure(self):
                self.index = 0

            def process(self, blob):
                assert blob['blob_index'] == self.index
                self.index += 1
                return blob

        pipe = kp.Pipeline()
        pipe.attach(InfinitePump)
        pipe.attach(BlobIndexer)
        pipe.attach(Observer)
        pipe.drain(4)


class TestLocalDBService(TestCase):
    def test_create_table(self):
        fobj = tempfile.NamedTemporaryFile(delete=True)
        dbs = LocalDBService(filename=fobj.name)
        dbs.create_table("foo", ["a", "b"], ["INT", "TEXT"])
        assert dbs.table_exists("foo")

    def test_create_table_does_not_overwrite_by_default(self):
        fobj = tempfile.NamedTemporaryFile(delete=True)
        dbs = LocalDBService(filename=fobj.name)
        dbs.create_table("foo", ["a", "b"], ["INT", "TEXT"])
        with self.assertRaises(sqlite3.OperationalError):
            dbs.create_table("foo", ["a", "b"], ["INT", "TEXT"])

    def test_create_table_allows_overwrite(self):
        fobj = tempfile.NamedTemporaryFile(delete=True)
        dbs = LocalDBService(filename=fobj.name)
        dbs.create_table("foo", ["a", "b"], ["INT", "TEXT"])
        dbs.create_table("foo", ["a", "b"], ["INT", "TEXT"], overwrite=True)

    def test_insert_row(self):
        fobj = tempfile.NamedTemporaryFile(delete=True)
        dbs = LocalDBService(filename=fobj.name)
        dbs.create_table("foo", ["a", "b"], ["INT", "TEXT"])

        dbs.insert_row("foo", ["a", "b"], (23, "42"))
        dbs.insert_row("foo", ["a", "b"], (5, "hello"))

        cur = dbs.connection.cursor()
        cur.execute("SELECT * FROM foo")
        data = cur.fetchall()
        assert 2 == len(data)
        assert 23 == data[0][0]
        assert "42" == data[0][1]
        assert 5 == data[1][0]
        assert "hello" == data[1][1]



# Filename: test_config.py
"""
Test suite for configuration related functions and classes.

"""

from io import StringIO

from km3pipe.testing import TestCase
from km3pipe.config import Config

__author__ = "Tamas Gal"
__copyright__ = "Copyright 2016, Tamas Gal and the KM3NeT collaboration."
__credits__ = []
__license__ = "MIT"
__maintainer__ = "Tamas Gal"
__email__ = "tgal@km3net.de"
__status__ = "Development"

CONFIGURATION = StringIO(
    "\n".join((
        "[DB]",
        "username=foo",
        "password=narf",
        "timeout=10",
    ))
)


class TestConfig(TestCase):
    def setUp(self):
        self.config = Config(None)
        self.config._read_from_file(CONFIGURATION)
        CONFIGURATION.seek(0, 0)

    def test_db_credentials(self):
        self.assertEqual('foo', self.config.db_credentials[0])
        self.assertEqual('narf', self.config.db_credentials[1])

    def test_check_for_updates_defaults_to_true(self):
        self.assertTrue(self.config.check_for_updates)

    def test_time_zone_defaults_to_utc(self):
        self.assertEqual('UTC', self.config.time_zone._tzname)

    def test_slack_token_raises_error_by_default(self):
        with self.assertRaises(ValueError):
            self.config.slack_token

    def test_get_retrieves_correct_value(self):
        self.assertEqual("foo", self.config.get("DB", "username"))

    def test_get_returns_none_if_section_not_found(self):
        self.assertTrue(self.config.get("a", "b") is None)

    def test_get_returns_none_if_option_not_found(self):
        self.assertTrue(self.config.get("DB", "a") is None)

    def test_get_returns_default_if_option_not_found(self):
        self.assertEqual('b', self.config.get("DB", "a", default='b'))

    def test_get_returns_float_if_option_is_numberlike(self):
        self.assertTrue(isinstance(self.config.get("DB", "timeout"), float))

    def test_create_irods_session_returns_none_if_irods_module_missing(self):
        session = self.config.create_irods_session()
        self.assertTrue(session is None)



# Filename: dataclasses.py
# pylint: disable=W0232,C0103,C0111
# vim:set ts=4 sts=4 sw=4 et syntax=python:
"""
Dataclasses for internal use. Heavily based on Numpy arrays.
"""
import numpy as np

__author__ = "Tamas Gal and Moritz Lotze"
__copyright__ = "Copyright 2016, Tamas Gal and the KM3NeT collaboration."
__credits__ = []
__license__ = "MIT"
__maintainer__ = "Moritz Lotze"
__email__ = "mlotze@km3net.de"
__status__ = "Development"
__all__ = ('TEMPLATES', )

TEMPLATES = {
    'Direction': {
        'dtype': np.dtype([
            ('dir_x', '<f4'),
            ('dir_y', '<f4'),
            ('dir_z', '<f4'),
        ]),
        'h5loc': None,
        'split_h5': False,
        'h5singleton': False,
    },
    'Position': {
        'dtype': np.dtype([
            ('pos_x', '<f4'),
            ('pos_y', '<f4'),
            ('pos_z', '<f4'),
        ]),
        'h5loc': None,
        'split_h5': False,
        'h5singleton': False,
    },
    'EventInfo': {
        'dtype': np.dtype([
            ('det_id', '<i4'),
            ('frame_index', '<u4'),
            ('livetime_sec', '<u8'),
            ('mc_id', '<i4'),
            ('mc_t', '<f8'),
            ('n_events_gen', '<u8'),
            ('n_files_gen', '<u8'),
            ('overlays', '<u4'),
            ('trigger_counter', '<u8'),
            ('trigger_mask', '<u8'),
            ('utc_nanoseconds', '<u8'),
            ('utc_seconds', '<u8'),
            ('weight_w1', '<f8'),
            ('weight_w2', '<f8'),
            ('weight_w3', '<f8'),
            ('run_id', '<u8'),
            ('group_id', '<u8'),
        ]),
        'h5loc': '/event_info',
        'split_h5': False,
        'h5singleton': False,
    },
    'TimesliceHits': {
        'dtype': np.dtype([
            ('channel_id', 'u1'),
            ('dom_id', '<u4'),
            ('time', '<f8'),
            ('tot', 'u1'),
            ('group_id', '<u4'),
        ]),
        'h5loc': '/timeslice_hits',
        'split_h5': True,
        'h5singleton': False,
    },
    'Hits': {
        'dtype': np.dtype([
            ('channel_id', 'u1'),
            ('dom_id', '<u4'),
            ('time', '<f8'),
            ('tot', 'u1'),
            ('triggered', '?'),
            ('group_id', '<u4'),
        ]),
        'h5loc': '/hits',
        'split_h5': True,
        'h5singleton': False,
    },
    'CalibHits': {
        'dtype': np.dtype([
            ('channel_id', 'u1'),
            ('dir_x', '<f4'),
            ('dir_y', '<f4'),
            ('dir_z', '<f4'),
            ('dom_id', '<u4'),
            ('du', 'u1'),
            ('floor', 'u1'),
            ('pos_x', '<f4'),
            ('pos_y', '<f4'),
            ('pos_z', '<f4'),
            ('t0', '<f4'),
            ('time', '<f8'),
            ('tot', 'u1'),
            ('triggered', '?'),
            ('group_id', '<u4'),
        ]),
        'h5loc': '/hits',
        'split_h5': True,
        'h5singleton': False,
    },
    'McHits': {
        'dtype': np.dtype([
            ('a', '<f4'),
            ('origin', '<u4'),
            ('pmt_id', '<u4'),
            ('time', '<f8'),
            ('group_id', '<u4'),
        ]),
        'h5loc': '/mc_hits',
        'split_h5': True,
        'h5singleton': False,
    },
    'CalibMcHits': {
        'dtype': np.dtype([
            ('a', '<f4'),
            ('dir_x', '<f4'),
            ('dir_y', '<f4'),
            ('dir_z', '<f4'),
            ('origin', '<u4'),
            ('pmt_id', '<u4'),
            ('pos_x', '<f4'),
            ('pos_y', '<f4'),
            ('pos_z', '<f4'),
            ('time', '<f8'),
            ('group_id', '<u4'),
        ]),
        'h5loc': '/mc_hits',
        'split_h5': True,
        'h5singleton': False,
    },
    'Tracks': {
        'dtype': np.dtype([
            ('bjorkeny', '<f8'),
            ('dir_x', '<f8'),
            ('dir_y', '<f8'),
            ('dir_z', '<f8'),
            ('energy', '<f8'),
            ('id', '<u4'),
            ('interaction_channel', '<u4'),
            ('is_cc', '<u4'),    # TODO: consider bool ('?') for slicing
            ('length', '<f8'),
            ('pos_x', '<f8'),
            ('pos_y', '<f8'),
            ('pos_z', '<f8'),
            ('time', '<i4'),
            ('type', '<i4'),
            ('group_id', '<u4'),
        ]),
        'h5loc': '/tracks',
        'split_h5': False,
        'h5singleton': False,
    },
    'McTracks': {
        'dtype': np.dtype([
            ('bjorkeny', '<f8'),
            ('dir_x', '<f8'),
            ('dir_y', '<f8'),
            ('dir_z', '<f8'),
            ('energy', '<f8'),
            ('id', '<u4'),
            ('interaction_channel', '<u4'),
            ('is_cc', '<u4'),    # TODO: consider bool ('?') for slicing
            ('length', '<f8'),
            ('pos_x', '<f8'),
            ('pos_y', '<f8'),
            ('pos_z', '<f8'),
            ('time', '<i4'),
            ('type', '<i4'),
            ('group_id', '<u4'),
        ]),
        'h5loc': '/mc_tracks',
        'split_h5': False,
        'h5singleton': False,
    },
    'SummaryFrameInfo': {
        'dtype': np.dtype([
            ('dom_id', '<u4'),
            ('fifo_status', '<u4'),
            ('frame_id', '<u4'),
            ('frame_index', '<u4'),
            ('has_udp_trailer', '<u4'),
            ('high_rate_veto', '<u4'),
            ('max_sequence_number', '<u4'),
            ('n_packets', '<u4'),
            ('slice_id', '<u4'),
            ('utc_nanoseconds', '<u4'),
            ('utc_seconds', '<u4'),
            ('white_rabbit_status', '<u4'),
        ]),
        'h5loc': '/summary_slice_info',
        'split_h5': False,
        'h5singleton': False,
    },
    'SummarysliceInfo': {
        'dtype': np.dtype([
            ('frame_index', '<u4'),
            ('slice_id', '<u4'),
            ('timestamp', '<u4'),
            ('nanoseconds', '<u4'),
            ('n_frames', '<u4'),
        ]),
        'h5loc': '/todo',
        'split_h5': False,
        'h5singleton': False,
    },
    'TimesliceInfo': {
        'dtype': np.dtype([
            ('frame_index', '<u4'),
            ('slice_id', '<u4'),
            ('timestamp', '<u4'),
            ('nanoseconds', '<u4'),
            ('n_frames', '<u4'),
        ]),
        'h5loc': '/timeslice_info',
        'split_h5': False,
        'h5singleton': False,
    },
    'SummaryframeSeries': {
        'dtype': np.dtype([
            ('dom_id', '<u4'),
            ('max_sequence_number', '<u4'),
            ('n_received_packets', '<u4'),
            ('group_id', '<u4'),
        ]),
        'h5loc': '/todo',
        'split_h5': False,
        'h5singleton': False,
    },
    'TimesliceFrameInfo': {
        'dtype': np.dtype([
            ('det_id', '<i4'),
            ('run_id', '<u8'),
            ('sqnr', '<u8'),
            ('timestamp', '<u4'),
            ('nanoseconds', '<u4'),
            ('dom_id', '<u4'),
            ('dataqueue_status', '<u4'),
            ('dom_status', '<u4'),
            ('n_hits', '<u4'),
        ]),
        'h5loc': '/todo',
        'split_h5': False,
        'h5singleton': False,
    },
    'SummaryFrameSeries': {
        'dtype': np.dtype([
            ('dom_id', '<u4'),
            ('max_sequence_number', '<u4'),
            ('n_received_packets', '<u4'),
            ('group_id', '<u4'),
        ]),
        'h5loc': '/summary_frame_series',
        'split_h5': False,
        'h5singleton': False,
    }
}



#!/usr/bin/env python

__author__ = 'tamasgal'

import km3pipe as kp


class DummyPump(kp.Module):
    """A pump demonstration with a dummy list as data."""
    def configure(self):
        self.data = [{'nr': 1}, {'nr': 2}]
        self.blobs = self.blob_generator()

    def process(self, blob):
        return next(self.blobs)

    def blob_generator(self):
        """Create a blob generator."""
        for blob in self.data:
            yield blob


class Foo(kp.Module):
    """A dummy module with optional and required parameters"""
    def configure(self):
        self.foo = self.get('foo', default='default_foo')    # optional
        self.bar = self.get('bar', default=23)    # optional
        self.baz = self.require('baz')    # required
        self.i = 0

    def process(self, blob):
        print("This is the current blob: " + str(blob))
        self.i += 1
        blob['foo_entry'] = self.foo
        return blob

    def finish(self):
        print("My process() method was called {} times.".format(self.i))


def moo(blob):
    """A simple function to attach"""
    blob['moo_entry'] = 42
    return blob


class PrintBlob(kp.Module):
    def process(self, blob):
        print(blob)
        return blob


pipe = kp.Pipeline()
pipe.attach(DummyPump, 'the_pump')
pipe.attach(Foo, bar='dummybar', baz=69)
pipe.attach(moo)
pipe.attach(PrintBlob)
pipe.drain()



# -*- coding: utf-8 -*-
"""

===========
Convex Hull
===========

Convex hull of a set of points, representing Dom x-y positions.

Derived from ``scipy.spatial.qhull.pyx``.
"""

import numpy as np
from scipy.spatial import ConvexHull
import matplotlib.pyplot as plt
from mpl_toolkits.mplot3d import Axes3D    # noqa
from km3net_testdata import data_path

from km3pipe.hardware import Detector
from km3pipe.math import Polygon
import km3pipe.style
km3pipe.style.use("km3pipe")

detx = data_path('detx/orca_115strings_av23min20mhorizontal_18OMs_alt9mvertical_v1.detx')
detector = Detector(detx)
xy = detector.xy_positions
hull = ConvexHull(xy)

##############################################################################
# Plot it:

plt.plot(xy[:, 0], xy[:, 1], 'o')
for simplex in hull.simplices:
    plt.plot(xy[simplex, 0], xy[simplex, 1], 'k-')

##############################################################################
# We could also have directly used the vertices of the hull, which
# for 2-D are guaranteed to be in counterclockwise order:

plt.plot(xy[hull.vertices, 0], xy[hull.vertices, 1], 'r--', lw=2)
plt.plot(xy[hull.vertices[0], 0], xy[hull.vertices[0], 1], 'ro')
plt.show()

##############################################################################
# Now let's draw a polygon inside, and see which points are contained.

poly_vertices = np.array([
    (-60, 120),
    (80, 120),
    (110, 60),
    (110, -30),
    (70, -110),
    (-70, -110),
    (-90, -70),
    (-90, 60),
])
poly = Polygon(poly_vertices)
contain_mask = poly.contains(xy)

##############################################################################
# and color them accordingly

plt.clf()
plt.plot(xy[contain_mask, 0], xy[contain_mask, 1], 'yo')
plt.plot(xy[~contain_mask, 0], xy[~contain_mask, 1], 'bo')
plt.plot(poly_vertices[:, 0], poly_vertices[:, 1], 'k-')
plt.show()

##############################################################################
# plot the same in 3D because why not?

plt.clf()
fig = plt.figure()
ax = fig.add_subplot(111, projection='3d')
ax.scatter(xy[:, 0], xy[:, 1], 90, 'yo')
ax.scatter(xy[:, 0], xy[:, 1], -90, 'bo')
ax.plot(poly_vertices[:, 0], poly_vertices[:, 1], 90, 'k-')
for simplex in hull.simplices:
    ax.plot(xy[simplex, 0], xy[simplex, 1], -90, 'k-')
plt.show()



__author__ = 'tamasgal'

import matplotlib.pyplot as plt

from km3pipe import Pipeline, Module
from km3pipe.io import CLBPump
from km3modules.common import StatusBar


class TOTHisto(Module):
    def configure(self):
        self.tots = []

    def process(self, blob):
        for pmt_data in blob['PMTData']:
            self.tots.append(pmt_data.tot)
        return blob

    def finish(self):
        plt.hist(self.tots, 80)
        plt.xlabel("ToT [ns]")
        plt.ylabel('count')
        plt.show()


class PrintCLBHeader(Module):
    def process(self, blob):
        print(blob['CLBHeader'])
        return blob


pipeline = Pipeline()
pipeline.attach(
    CLBPump, filename='/Users/tamasgal/Data/KM3NeT/du1-clb/DOM2_run23.dat'
)
pipeline.attach(StatusBar)
pipeline.attach(PrintCLBHeader)
pipeline.attach(TOTHisto)
pipeline.drain(30)



# Filename: test_math.py
# pylint: disable=locally-disabled,C0111,R0904,C0103

import numpy as np
from numpy.testing import (assert_almost_equal, assert_allclose)
import pytest

from km3pipe.testing import TestCase
from km3pipe.math import (
    angle_between, dist, pld3, com, zenith, azimuth, Polygon, IrregularPrism,
    rotation_matrix, SparseCone, space_angle, hsin, phi, theta, unit_vector,
    innerprod_1d, log_b, qeuler, qrot, qrot_yaw, intersect_3d
)

__author__ = ["Tamas Gal", "Moritz Lotze"]
__copyright__ = "Copyright 2016, KM3Pipe devs and the KM3NeT collaboration."
__credits__ = ["Thomas Heid"]
__license__ = "MIT"
__maintainer__ = ["Tamas Gal", "Moritz Lotze"]
__email__ = "tgal@km3net.de"
__status__ = "Development"


class TestMath(TestCase):
    def setUp(self):
        # self.vecs = np.array([[0., 1., 5.],
        #                       [1., 1., 4.],
        #                       [2., 1., 3.],
        #                       [3., 1., 2.],
        #                       [4., 1., 1.]])
        # self.v = (1, 2, 3)
        self.v = np.array([0.26726124, 0.53452248, 0.80178373])
        self.vecs = np.array([[0., 0.19611614, 0.98058068],
                              [0.23570226, 0.23570226, 0.94280904],
                              [0.53452248, 0.26726124, 0.80178373],
                              [0.80178373, 0.26726124, 0.53452248],
                              [0.94280904, 0.23570226, 0.23570226]])

    def test_phi(self):
        print(phi((1, 0, 0)))
        assert_almost_equal(0, phi((1, 0, 0)))
        assert_almost_equal(np.pi, phi((-1, 0, 0)))
        assert_almost_equal(np.pi / 2, phi((0, 1, 0)))
        assert_almost_equal(np.pi / 2 * 3, phi((0, -1, 0)))
        assert_almost_equal(np.pi / 2 * 3, phi((0, -1, 0)))
        assert_almost_equal(0, phi((0, 0, 0)))
        assert_almost_equal(phi(self.v), 1.10714872)
        assert_almost_equal(
            phi(self.vecs),
            np.array([
                1.57079633, 0.78539816, 0.46364761, 0.32175055, 0.24497866
            ])
        )

    def test_zenith(self):
        assert_allclose(np.pi, zenith((0, 0, 1)))
        assert_allclose(0, zenith((0, 0, -1)))
        assert_allclose(np.pi / 2, zenith((0, 1, 0)))
        assert_allclose(np.pi / 2, zenith((0, -1, 0)))
        assert_allclose(np.pi / 4 * 3, zenith((0, 1, 1)))
        assert_allclose(np.pi / 4 * 3, zenith((0, -1, 1)))
        assert_almost_equal(zenith(self.v), 2.5010703409103687)
        assert_allclose(
            zenith(self.vecs),
            np.array([
                2.94419709, 2.80175574, 2.50107034, 2.13473897, 1.80873745
            ])
        )

    def test_azimuth(self):
        self.assertTrue(np.allclose(np.pi, azimuth((1, 0, 0))))
        self.assertTrue(np.allclose(0, azimuth((-1, 0, 0))))

        print(azimuth((0, 1, 0)))
        print(azimuth((0, -1, 0)))
        print(azimuth((0, 0, 0)))
        print(azimuth(self.v))
        print(azimuth(self.vecs))
        self.assertTrue(np.allclose(np.pi / 2 * 3, azimuth((0, 1, 0))))
        self.assertTrue(np.allclose(np.pi / 2, azimuth((0, -1, 0))))
        self.assertTrue(np.allclose(np.pi, azimuth((0, 0, 0))))
        self.assertTrue(np.allclose(azimuth(self.v), 4.24874137138))
        self.assertTrue(
            np.allclose(
                azimuth(self.vecs),
                np.array([
                    4.71238898, 3.92699082, 3.60524026, 3.46334321, 3.38657132
                ])
            )
        )

    def test_theta(self):
        print(theta((0, 0, -1)))
        print(theta((0, 0, 1)))
        print(theta((0, 1, 0)))
        print(theta((0, -1, 0)))
        print(theta((0, 1, 1)))
        print(theta((0, -1, 1)))
        print(theta(self.v))
        print(theta(self.vecs))
        self.assertTrue(np.allclose(0, theta((0, 0, 1))))
        self.assertTrue(np.allclose(np.pi, theta((0, 0, -1))))
        self.assertTrue(np.allclose(np.pi / 2, theta((0, 1, 0))))
        self.assertTrue(np.allclose(np.pi / 2, theta((0, -1, 0))))
        self.assertTrue(np.allclose(0, theta((0, 1, 1))))
        self.assertTrue(np.allclose(0, theta((0, -1, 1))))
        self.assertTrue(np.allclose(theta(self.v), 0.64052231))
        self.assertTrue(
            np.allclose(
                theta(self.vecs),
                np.array([
                    0.19739554, 0.33983691, 0.64052231, 1.00685369, 1.3328552
                ])
            )
        )

    def test_unit_vector(self):
        v1 = (1, 0, 0)
        v2 = (1, 1, 0)
        v3 = (-1, 2, 0)
        assert np.allclose(v1, unit_vector(v1))
        assert np.allclose(np.array(v2) / np.sqrt(2), unit_vector(v2))
        assert np.allclose(np.array(v3) / np.sqrt(5), unit_vector(v3))

    def test_innerprod_1d(self):
        v1 = [[1, 2, 3], [4, 5, 6]]
        v2 = [[11, 22, 33], [44, 55, 66]]
        assert np.allclose([14, 77], innerprod_1d(v1, v2))

    def test_angle_between(self):
        v1 = (1, 0, 0)
        v2 = (0, 1, 0)
        v3 = (-1, 0, 0)
        self.assertAlmostEqual(0, angle_between(v1, v1))
        self.assertAlmostEqual(np.pi / 2, angle_between(v1, v2))
        self.assertAlmostEqual(np.pi, angle_between(v1, v3))
        self.assertAlmostEqual(angle_between(self.v, v1), 1.3002465638163236)
        self.assertAlmostEqual(angle_between(self.v, v2), 1.0068536854342678)
        self.assertAlmostEqual(angle_between(self.v, v3), 1.8413460897734695)
        self.assertTrue(
            np.allclose(
                angle_between(self.vecs, v1),
                np.array([
                    1.57079633, 1.3328552, 1.0068537, 0.64052231, 0.33983691
                ])
            )
        )
        self.assertTrue(
            np.allclose(
                angle_between(self.vecs, v2),
                np.array([
                    1.37340077, 1.3328552, 1.3002466, 1.30024656, 1.3328552
                ])
            )
        )
        self.assertTrue(
            np.allclose(
                angle_between(self.vecs, v3),
                np.array([
                    1.57079633, 1.80873745, 2.13473897, 2.50107034, 2.80175574
                ])
            )
        )

    def test_angle_between_returns_nan_for_zero_length_vectors(self):
        v1 = (0, 0, 0)
        v2 = (1, 0, 0)
        with pytest.warns(RuntimeWarning):
            self.assertTrue(np.isnan(angle_between(v1, v2)))

    def test_space_angle(self):
        p1 = (np.pi / 2, np.pi)
        p2 = (np.pi, 0)
        self.assertAlmostEqual(
            space_angle(p1[0], p2[0], p1[1], p2[1]), 1.57079632679489
        )
        p3 = (0, np.pi)
        p4 = (np.pi / 2, 0)
        self.assertAlmostEqual(
            space_angle(p3[0], p4[0], p3[1], p4[1]), 1.57079632679489
        )

    def test_hsin(self):
        assert np.all(hsin((np.pi, 0)) == (1, 0))
        self.assertAlmostEqual(hsin(np.pi / 2), 0.5)

    def test_pld3(self):
        p1 = np.array((0, 0, 0))
        p2 = np.array((0, 0, 1))
        d2 = np.array((0, 1, 0))
        self.assertAlmostEqual(1, pld3(p1, p2, d2))
        p1 = np.array((0, 0, 0))
        p2 = np.array((0, 0, 2))
        d2 = np.array((0, 1, 0))
        self.assertAlmostEqual(2, pld3(p1, p2, d2))
        p1 = np.array((0, 0, 0))
        p2 = np.array((0, 0, 0))
        d2 = np.array((0, 1, 0))
        self.assertAlmostEqual(0, pld3(p1, p2, d2))
        p1 = np.array((1, 2, 3))
        p2 = np.array((4, 5, 6))
        d2 = np.array((7, 8, 9))
        self.assertAlmostEqual(0.5275893, pld3(p1, p2, d2))
        p1 = np.array((0, 0, 2))
        p2 = np.array((-100, 0, -100))
        d2 = np.array((1, 0, 1))
        self.assertAlmostEqual(1.4142136, pld3(p1, p2, d2))
        p1 = np.array([183., -311., 351.96083871])
        p2 = np.array([40.256, -639.888, 921.93])
        d2 = np.array([0.185998, 0.476123, -0.859483])
        self.assertAlmostEqual(21.25456308, pld3(p1, p2, d2))

    def test_com(self):
        center_of_mass = com(((1, 2, 3), (4, 5, 6), (7, 8, 9), (10, 11, 12)))
        self.assertEqual((5.5, 6.5, 7.5), tuple(center_of_mass))
        center_of_mass = com(((1, 2, 3), (4, 5, 6), (7, 8, 9)),
                             masses=(1, 0, 0))
        self.assertEqual((1, 2, 3), tuple(center_of_mass))
        center_of_mass = com(((1, 1, 1), (0, 0, 0)))
        self.assertEqual((0.5, 0.5, 0.5), tuple(center_of_mass))


class TestShapes(TestCase):
    def setUp(self):
        self.poly = [
            (-60, 120),
            (80, 120),
            (110, 60),
            (110, -30),
            (70, -110),
            (-70, -110),
            (-90, -70),
            (-90, 60),
        ]

    def test_poly_containment(self):
        polygon = Polygon(self.poly)
        point_in = (-40, -40)
        point_out = (-140, -140)
        points = [
            (-40, -40),
            (-140, -140),
            (40, -140),
        ]
        assert np.all(polygon.contains(point_in))
        assert not np.any(polygon.contains(point_out))
        assert np.all(polygon.contains(points) == [True, False, False])

    def test_poly_xy(self):
        polygon = Polygon(self.poly)
        x = (-40, -140, 40)
        y = (-40, -140, -140)
        assert np.all(polygon.contains_xy(x, y) == [True, False, False])

    def test_prism_contained(self):
        z = (-90, 90)
        prism = IrregularPrism(self.poly, z[0], z[1])
        points = [
            (0, 1, 2),
            (-100, 20, 10),
            (10, 90, 10),
        ]
        assert np.all(prism.contains(points) == [True, False, True])

    def test_prism_contained_xyz(self):
        z = (-90, 90)
        prism = IrregularPrism(self.poly, z[0], z[1])
        x = (0, -100, 10)
        y = (1, 20, 90)
        z = (2, 10, 10)
        assert np.all(prism.contains_xyz(x, y, z) == [True, False, True])


class TestRotation(TestCase):
    def test_rotmat(self):
        v = [3, 5, 0]
        axis = [4, 4, 1]
        theta = 1.2
        newvec = np.dot(rotation_matrix(axis, theta), v)
        self.assertTrue(
            np.allclose(
                newvec, np.array([2.74911638, 4.77180932, 1.91629719])
            )
        )

    def test_cone(self):
        spike = [1, 1, 0]
        bottom = [0, 2, 0]
        angle = np.pi / 4
        n_angles = 20
        cone = SparseCone(spike, bottom, angle)
        circ_samp = cone.sample_circle(n_angles=n_angles)
        axis_samp = cone.sample_axis
        samp = cone.sample(n_angles)
        assert len(circ_samp) == n_angles
        assert len(axis_samp) == 2
        assert len(samp) == len(circ_samp) + 2


class TestLog(TestCase):
    def test_val(self):
        assert_allclose(log_b(5, 2), np.log2(5))
        assert_allclose(log_b(5, 10), np.log10(5))
        assert_allclose(log_b(5, np.e), np.log(5))


class TestQeuler(TestCase):
    def test_conversion_of_yaw(self):
        assert np.allclose([1, 0, 0, 0], qeuler(0, 0, 0))
        assert np.allclose([0.7071, 0, 0, 0.7071], qeuler(90, 0, 0))
        assert np.allclose([0, 0, 0, 1], qeuler(180, 0, 0))
        assert np.allclose([-0.7071, 0, 0, 0.7071], qeuler(270, 0, 0))
        assert np.allclose([-1, 0, 0, 0], qeuler(360, 0, 0))

    def test_conversion_of_pitch(self):
        assert np.allclose([0.92388, 0, 0.38268, 0], qeuler(0, 45, 0))
        assert np.allclose([0.92388, 0, -0.38268, 0], qeuler(0, -45, 0))
        assert np.allclose([0.7071, 0, 0.7071, 0], qeuler(0, 90, 0))
        assert np.allclose([0.8660254, 0, 0.5, 0], qeuler(0, 60, 0))
        assert np.allclose([-0.96592583, 0, -0.25881905, 0], qeuler(0, 390, 0))

    def test_conversion_of_roll(self):
        assert np.allclose([0.92388, 0.38268, 0, 0], qeuler(0, 0, 45))
        assert np.allclose([0.92388, -0.38268, 0, 0], qeuler(0, 0, -45))
        assert np.allclose([0.70710, 0.70710, 0, 0], qeuler(0, 0, 90))
        assert np.allclose([0.86602, 0.5, 0, 0], qeuler(0, 0, 60))
        assert np.allclose([-0.96592583, -0.25881905, 0, 0], qeuler(0, 0, 390))

    def test_mixed_conversion(self):
        assert np.allclose([0.999471, 0.02601972, 0.01767416, 0.00826538],
                           qeuler(1, 2, 3))
        assert np.allclose([0.94371436, 0.26853582, -0.14487813, 0.12767944],
                           qeuler(10, -20, 30))
        assert np.allclose([-0.16575384, -0.69624819, 0.05479592, -0.69624819],
                           qeuler(-999, 999, -999))


class TestQrot(TestCase):
    def test_rotation_of_x_vector(self):
        assert np.allclose([0, 1, 0], qrot([1, 0, 0], qeuler(90, 0, 0)))
        assert np.allclose([-1, 0, 0], qrot([1, 0, 0], qeuler(180, 0, 0)))
        assert np.allclose([-1, 0, 0], qrot([1, 0, 0], qeuler(180, 0, -45)))
        assert np.allclose([0, 0, -1], qrot([1, 0, 0], qeuler(180, 90, 45)))

    def test_rotation_of_y_vector(self):
        assert np.allclose([-1, 0, 0], qrot([0, 1, 0], qeuler(90, 0, 0)))
        assert np.allclose([0, -1, 0], qrot([0, 1, 0], qeuler(180, 0, 0)))
        assert np.allclose([0, -0.70710, -0.70710],
                           qrot([0, 1, 0], qeuler(180, 0, -45)))
        assert np.allclose([-0.70710, -0.70710, 0],
                           qrot([0, 1, 0], qeuler(180, 90, 45)))

    def test_rotation_of_z_vector(self):
        assert np.allclose([0, 0, 1], qrot([0, 0, 1], qeuler(90, 0, 0)))
        assert np.allclose([0, 0, 1], qrot([0, 0, 1], qeuler(180, 0, 0)))
        assert np.allclose([0, -0.70710, 0.70710],
                           qrot([0, 0, 1], qeuler(180, 0, -45)))
        assert np.allclose([-0.70710, 0.70710, 0],
                           qrot([0, 0, 1], qeuler(180, 90, 45)))

    def test_mixed_rotation(self):
        assert np.allclose([1, 2, 3], qrot([1, 2, 3], qeuler(0, 0, 0)))
        assert np.allclose([0, -1.414213, 0],
                           qrot([0, 1, -1], qeuler(180, 90, 45)))
        assert np.allclose([-1.41421356, 0, -1],
                           qrot([1, 1, 1], qeuler(180, 90, 45)))
        assert np.allclose([-14.1421356, 0, -10],
                           qrot([10, 10, 10], qeuler(180, 90, 45)))


class TestQrotYaw(TestCase):
    def test_call_with_list(self):
        qrot_yaw([1, 2, 3], 1)

    def test_no_rotation(self):
        vec = (1, 0, 0)
        vec_rot = qrot_yaw(vec, 0)
        assert np.allclose([1, 0, 0], vec_rot)

    def test_a_rotation_of_90(self):
        vec = (1, 0, 0)
        vec_rot = qrot_yaw(vec, 90)
        assert np.allclose([0, 1, 0], vec_rot)

    def test_a_rotation_of_180(self):
        vec = (1, 0, 0)
        vec_rot = qrot_yaw(vec, 180)
        assert np.allclose([-1, 0, 0], vec_rot)

    def test_a_full_rotation(self):
        vec = (1, 0, 0)
        vec_rot = qrot_yaw(vec, 360)
        assert np.allclose([1, 0, 0], vec_rot)

    def test_a_rotation_of_45(self):
        vec = (1, 0, 0)
        vec_rot = qrot_yaw(vec, 45)
        assert np.allclose([0.7071, 0.7071, 0], vec_rot)


class TestIntersect3D(TestCase):
    def test_intersection_at_zero(self):
        p1 = np.array([(1, 0, 0), (0, 0, 1)])
        p2 = -p1
        intersection = intersect_3d(p1, p2)
        assert np.allclose([0, 0, 0], intersection)

    def test_intersection_of_multiple_lines_with_same_endpoints(self):
        p1 = np.array([(1, 2, 3), (4, 5, 6), (7, 8, 9)])
        p2 = np.array([(4, 4, 4), (4, 4, 4), (4, 4, 4)])
        intersection = intersect_3d(p1, p2)
        assert np.allclose([4, 4, 4], intersection)

    def test_intersection_of_multiple_lines_with_target(self):
        p1 = np.array([(1, 2, 3), (4, 5, 6), (7, 8, 9)])
        target = np.array([23, 5, 42])
        p2 = 2 * target - p1
        intersection = intersect_3d(p1, p2)
        assert np.allclose(target, intersection)

    def test_another_intersection(self):
        p1 = np.array([(1, 10, 0), (0, 10, 1)])
        p2 = np.array([(-1, 10, 0), (0, 10, -1)])
        intersection = intersect_3d(p1, p2)
        assert np.allclose([0, 10, 0], intersection)


class TestDist(TestCase):
    def test_dist_between_two_2D_points(self):
        self.assertAlmostEqual(1, dist(np.array([0, 0]), np.array([1, 0])))
        self.assertAlmostEqual(
            np.sqrt(2), dist(np.array([0, 1]), np.array([1, 0]))
        )
        self.assertAlmostEqual(
            2 * np.sqrt(2), dist(np.array([1, 2]), np.array([3, 4]))
        )

    def test_dist_between_two_3D_points(self):
        self.assertAlmostEqual(
            1, dist(np.array([0, 0, 0]), np.array([1, 0, 0]))
        )
        self.assertAlmostEqual(
            np.sqrt(2), dist(np.array([0, 1, 0]), np.array([1, 0, 0]))
        )
        self.assertAlmostEqual(
            2, dist(np.array([0, 0, 2]), np.array([0, 0, 0]))
        )
        self.assertAlmostEqual(
            5.1961524, dist(np.array([1, 2, 3]), np.array([4, 5, 6]))
        )

    def test_dist_to_many_points(self):
        assert np.allclose([1, 1, 0, 1.73205081],
                           dist(
                               np.array([0, 0, 0]),
                               np.array([[0, 0, 1], [0, 0, 1], [0, 0, 0],
                                         [1, 1, 1]]),
                               axis=1
                           ))



# Filename: test_evt.py
# pylint: disable=locally-disabled,C0111,R0904,C0301,C0103,W0212

from io import StringIO
from os.path import join, dirname

import numpy as np

from km3pipe.testing import TestCase, skip, data_path
from km3pipe.io.evt import EvtPump, EVT_PARSERS

__author__ = "Tamas Gal"
__copyright__ = "Copyright 2016, Tamas Gal and the KM3NeT collaboration."
__credits__ = []
__license__ = "MIT"
__maintainer__ = "Tamas Gal"
__email__ = "tgal@km3net.de"
__status__ = "Development"


class TestEvtPump(TestCase):
    def setUp(self):
        self.valid_evt_header = "\n".join((
            "start_run: 1", "cut_nu: 0.100E+03 0.100E+09-0.100E+01 0.100E+01",
            "spectrum: -1.40", "physics: gSeaGen 4.1 180126  165142",
            "physics: GENIE 2.10.2 180126  165142", "end_event:",
            "start_event: 12 1",
            "track_in: 1 -389.951 213.427 516 -0.204562 -0.60399 -0.770293" +
            " 9.092 0 5 40.998", "hit: 1 44675 1 1170.59 5 2 1 1170.59",
            "end_event:", "start_event: 13 1",
            "track_in:  1 272.695 -105.613 516 -0.425451 -0.370522 -0.825654" +
            " 2431.47 0 5 -1380",
            "track_in:  2 272.348 -106.292 516 -0.425451 -0.370522 -0.825654" +
            " 24670.7 1.33 5 -1484",
            "track_in:  3 279.47 -134.999 516 -0.425451 -0.370522 -0.825654" +
            " 164.586 26.7 5 601.939", "hit: 1 20140 1 1140.06 5 1 1 1140.06",
            "hit: 2 20159 1 1177.14 5 1 1 1177.14",
            "hit: 3 20164 1 1178.19 5 1 1 1178.19",
            "hit: 4 20170 1 1177.23 5 1 1 1177.23",
            "hit: 5 20171 2 1177.25 5 1 2 1177.25", "end_event:",
            "start_event: 14 1",
            "track_in:  1 40.256 -639.888 516 0.185998 0.476123 -0.859483" +
            " 10016.1 0 5 -1668", "hit: 1 33788 1 2202.81 5 1 1 2202.81",
            "hit: 2 33801 1 2248.95 5 1 1 2248.95",
            "hit: 3 33814 1 2249.2 5 1 1 2249.2", "end_event:"
        ))
        self.no_evt_header = "\n".join((
            "start_event: 12 1",
            "track_in: 1 -389.951 213.427 516 -0.204562 -0.60399 -0.770293" +
            " 9.092 0 5 40.998", "hit: 1 44675 1 1170.59 5 2 1 1170.59",
            "end_event:", "start_event: 13 1",
            "track_in:  1 272.695 -105.613 516 -0.425451 -0.370522 -0.825654" +
            " 2431.47 0 5 -1380",
            "track_in:  2 272.348 -106.292 516 -0.425451 -0.370522 -0.825654" +
            " 24670.7 1.33 5 -1484",
            "track_in:  3 279.47 -134.999 516 -0.425451 -0.370522 -0.825654" +
            " 164.586 26.7 5 601.939", "hit: 1 20140 1 1140.06 5 1 1 1140.06",
            "hit: 2 20159 1 1177.14 5 1 1 1177.14",
            "hit: 3 20164 1 1178.19 5 1 1 1178.19",
            "hit: 4 20170 1 1177.23 5 1 1 1177.23",
            "hit: 5 20171 2 1177.25 5 1 2 1177.25", "end_event:",
            "start_event: 14 1",
            "track_in:  1 40.256 -639.888 516 0.185998 0.476123 -0.859483" +
            " 10016.1 0 5 -1668", "hit: 1 33788 1 2202.81 5 1 1 2202.81",
            "hit: 2 33801 1 2248.95 5 1 1 2248.95",
            "hit: 3 33814 1 2249.2 5 1 1 2249.2", "end_event:"
        ))
        self.corrupt_evt_header = "foo"
        self.corrupt_line = "\n".join(
            ("start_event: 1 1", "corrupt line", "end_event:")
        )

        self.pump = EvtPump(parsers=[])
        self.pump.blob_file = StringIO(self.valid_evt_header)

    def tearDown(self):
        self.pump.blob_file.close()

    def test_parse_header(self):
        raw_header = self.pump.extract_header()
        self.assertEqual([['1']], raw_header['start_run'])
        self.assertAlmostEqual(-1.4, float(raw_header['spectrum'][0][0]))
        self.assertAlmostEqual(1, float(raw_header['cut_nu'][0][2]))

    def test_header_entries_with_same_tag_are_put_in_lists(self):
        raw_header = self.pump.extract_header()
        self.assertAlmostEqual(2, len(raw_header['physics']))
        self.assertAlmostEqual(1, len(raw_header['spectrum']))
        assert 'gSeaGen' == raw_header['physics'][0][0]
        assert 'GENIE' == raw_header['physics'][1][0]


#    def test_incomplete_header_raises_value_error(self):
#        temp_file = StringIO(self.corrupt_evt_header)
#        pump = EvtPump()
#        pump.blob_file = temp_file
#        with self.assertRaises(ValueError):
#            pump.extract_header()
#        temp_file.close()

    def test_record_offset_saves_correct_offset(self):
        self.pump.blob_file = StringIO('a' * 42)
        offsets = [1, 4, 9, 12, 23]
        for offset in offsets:
            self.pump.blob_file.seek(0, 0)
            self.pump.blob_file.seek(offset, 0)
            self.pump._record_offset()
        self.assertListEqual(offsets, self.pump.event_offsets)

    def test_event_offset_is_at_first_event_after_parsing_header(self):
        self.pump.extract_header()
        self.assertEqual(161, self.pump.event_offsets[0])

    def test_rebuild_offsets(self):
        self.pump.extract_header()
        self.pump._cache_offsets()
        self.assertListEqual([161, 306, 773], self.pump.event_offsets)

    def test_rebuild_offsets_without_header(self):
        self.pump.blob_file = StringIO(self.no_evt_header)
        self.pump.extract_header()
        self.pump._cache_offsets()
        self.assertListEqual([0, 145, 612], self.pump.event_offsets)

    def test_cache_enabled_triggers_rebuild_offsets(self):
        self.pump.cache_enabled = True
        self.pump.prepare_blobs()
        self.assertEqual(3, len(self.pump.event_offsets))

    def test_cache_disabled_doesnt_trigger_cache_offsets(self):
        self.pump.cache_enabled = False
        self.pump.prepare_blobs()
        self.assertEqual(1, len(self.pump.event_offsets))

    def test_get_blob_triggers_cache_offsets_if_cache_disabled(self):
        "...and asking for not indexed event"
        self.pump.cache_enabled = False
        self.pump.prepare_blobs()
        self.assertEqual(1, len(self.pump.event_offsets))
        blob = self.pump.get_blob(2)
        self.assertTupleEqual((14, 1), blob['start_event'])
        self.assertEqual(3, len(self.pump.event_offsets))

    def test_get_blob_raises_index_error_for_wrong_index(self):
        self.pump.prepare_blobs()
        with self.assertRaises(IndexError):
            self.pump.get_blob(23)

    def test_get_blob_returns_correct_event_information(self):
        self.pump.prepare_blobs()
        blob = self.pump.get_blob(0)
        self.assertTrue('raw_header' in blob)
        self.assertEqual([['1']], blob['raw_header']['start_run'])
        self.assertTupleEqual((12, 1), blob['start_event'])
        # TODO: all the other stuff like hit, track etc.
        assert 'hit' in blob
        assert 'track_in' in blob
        assert np.allclose([[
            1.0, 44675.0, 1.0, 1170.59, 5.0, 2.0, 1.0, 1170.59
        ]], blob['hit'])
        blob = self.pump.get_blob(1)
        assert 5 == len(blob['hit'])
        assert np.allclose([
            3.0, 20164.0, 1.0, 1178.19, 5.0, 1.0, 1.0, 1178.19
        ], blob['hit'][2])

    def test_get_blob_returns_correct_events(self):
        self.pump.prepare_blobs()
        blob = self.pump.get_blob(0)
        self.assertTupleEqual((12, 1), blob['start_event'])
        blob = self.pump.get_blob(2)
        self.assertTupleEqual((14, 1), blob['start_event'])
        blob = self.pump.get_blob(1)
        self.assertTupleEqual((13, 1), blob['start_event'])

    def test_process_returns_correct_blobs(self):
        self.pump.prepare_blobs()
        blob = self.pump.process()
        self.assertTupleEqual((12, 1), blob['start_event'])
        blob = self.pump.process()
        self.assertTupleEqual((13, 1), blob['start_event'])
        blob = self.pump.process()
        self.assertTupleEqual((14, 1), blob['start_event'])

    def test_process_raises_stop_iteration_if_eof_reached(self):
        self.pump.prepare_blobs()
        self.pump.process()
        self.pump.process()
        self.pump.process()
        with self.assertRaises(StopIteration):
            self.pump.process()

    def test_pump_acts_as_iterator(self):
        self.pump.prepare_blobs()
        event_numbers = []
        for blob in self.pump:
            event_numbers.append(blob['start_event'][0])
        self.assertListEqual([12, 13, 14], event_numbers)

    def test_pump_has_len(self):
        self.pump.prepare_blobs()
        self.assertEqual(3, len(self.pump))

    def test_pump_get_item_returns_first_for_index_zero(self):
        self.pump.prepare_blobs()
        first_blob = self.pump[0]
        self.assertEqual(12, first_blob['start_event'][0])

    def test_pump_get_item_returns_correct_blob_for_index(self):
        self.pump.prepare_blobs()
        blob = self.pump[1]
        self.assertEqual(13, blob['start_event'][0])

    def test_pump_slice_generator(self):
        self.pump.prepare_blobs()
        blobs = self.pump[:]
        blobs = list(self.pump[1:3])
        self.assertEqual(2, len(blobs))
        self.assertEqual((13, 1), blobs[0]['start_event'])

    def test_create_blob_entry_for_line_ignores_corrupt_line(self):
        self.pump.blob_file = StringIO(self.corrupt_line)
        self.pump.extract_header()
        self.pump.prepare_blobs()
        self.pump.get_blob(0)

    def test_parsers_are_ignored_if_not_valid(self):
        self.pump = EvtPump(parsers=['a', 'b'])
        self.pump.blob_file = StringIO(self.valid_evt_header)
        assert 'a' not in self.pump.parsers
        assert 'b' not in self.pump.parsers

    def test_parsers_are_added(self):
        self.pump = EvtPump(parsers=['km3sim'])
        self.pump.blob_file = StringIO(self.valid_evt_header)
        assert EVT_PARSERS['km3sim'] in self.pump.parsers

    def test_custom_parser(self):
        def a_parser(blob):
            blob['foo'] = 23

        self.pump = EvtPump(parsers=[a_parser])
        self.pump.blob_file = StringIO(self.valid_evt_header)
        self.pump.extract_header()
        self.pump.prepare_blobs()
        blob = self.pump[0]

        assert 23 == blob['foo']

    def test_auto_parser_finds_all_physics_parsers(self):
        self.pump = EvtPump(parsers='auto')
        self.pump.blob_file = StringIO(self.valid_evt_header)
        self.pump.extract_header()
        assert EVT_PARSERS['gseagen'] in self.pump.parsers


class TestEvtFilePump(TestCase):
    def setUp(self):
        self.fname = data_path("evt/example_numuNC.evt")

    def test_pipe(self):
        pump = EvtPump(filename=self.fname)
        next(pump)
        pump.finish()


class TestCorsika(TestCase):
    def setUp(self):
        self.fname = data_path('evt/example_corant_propa.evt')

    def test_pipe(self):
        pump = EvtPump(filename=self.fname)
        next(pump)
        pump.finish()


class TestPropa(TestCase):
    def setUp(self):
        self.fname = data_path('evt/example_corant_propa.evt')
        self.fnames = []
        for i in [0, 1]:
            self.fnames.append(data_path('evt/example_corant_propa.evt'))

    def test_pipe(self):
        pump = EvtPump(filename=self.fname, parsers=['propa'])
        assert EVT_PARSERS['propa'] in pump.parsers
        blob = next(pump)
        assert 'start_event' in blob
        assert 'track_primary' in blob
        assert 'track_in' in blob
        pump.finish()

    def test_filenames(self):
        pump = EvtPump(filenames=self.fnames, parsers=['propa'])
        assert EVT_PARSERS['propa'] in pump.parsers
        blob = next(pump)
        assert 'start_event' in blob
        assert 'track_primary' in blob
        assert 'track_in' in blob
        pump.finish()

    @skip
    def test_auto_parser(self):
        pump = EvtPump(filename=self.fname)
        assert EVT_PARSERS['propa'] in pump.parsers
        blob = next(pump)
        assert 'start_event' in blob
        assert 'track_primary' in blob
        assert 'Muon' in blob
        assert 'MuonMultiplicity' in blob
        assert 'Neutrino' in blob
        assert 'NeutrinoMultiplicity' in blob
        assert 'Weights' in blob
        assert 'Primary' in blob
        pump.finish()


class TestKM3Sim(TestCase):
    def setUp(self):
        self.fname = data_path('evt/KM3Sim.evt')

    def test_pipe(self):
        pump = EvtPump(filename=self.fname, parsers=['km3sim'])
        assert EVT_PARSERS['km3sim'] in pump.parsers
        next(pump)
        pump.finish()

    def test_hits(self):
        pump = EvtPump(filename=self.fname, parsers=['km3sim'])
        blob = pump[0]
        hits = blob['KM3SimHits']
        assert 4 == len(hits)
        assert 195749 == hits[0].pmt_id

    def test_neutrino(self):
        pump = EvtPump(filename=self.fname, parsers=['gseagen', 'km3sim'])
        blob = pump[0]
        EVT_PARSERS['gseagen'](blob)
        neutrino = blob['Neutrinos'][0]
        self.assertAlmostEqual(0.10066, neutrino.energy)


class TestParserDetection(TestCase):
    def test_parsers_are_automatically_detected(self):
        pass



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# vim: ts=4 sw=4 et
# Author: Tamas Gal <tgal@km3net.de>
# License: MIT
"""
==========
MSG reader
==========

This application demonstrates how to access MSGs from Ligier.

"""

import io
import km3pipe as kp

__author__ = "Tamas Gal"
__email__ = "tgal@km3net.de"


def print_msg(blob):
    data = io.BytesIO(blob['CHData'])
    print(data)
    return blob


pipe = kp.Pipeline()
pipe.attach(kp.io.CHPump, host='localhost', tags='MSG')
pipe.attach(print_msg)
pipe.drain()



# Filename: test_hardware.py
# pylint: disable=C0111,C0103,R0904
"""
Detector description (detx format v5)

global_det_id ndoms
dom_id line_id floor_id npmts
 pmt_id_global x y z dx dy dz t0
 pmt_id_global x y z dx dy dz t0
 ...
 pmt_id_global x y z dx dy dz t0
dom_id line_id floor_id npmts
 ...

"""

from copy import deepcopy
from os.path import join, dirname
from io import StringIO

import numpy as np

from km3pipe.testing import TestCase, data_path
from km3pipe.hardware import Detector, PMT
from km3pipe.math import qrot_yaw

__author__ = "Tamas Gal"
__copyright__ = "Copyright 2016, Tamas Gal and the KM3NeT collaboration."
__credits__ = []
__license__ = "MIT"
__maintainer__ = "Tamas Gal"
__email__ = "tgal@km3net.de"
__status__ = "Development"

EXAMPLE_DETX = StringIO(
    "\n".join((
        "1 3",
        "1 1 1 3",
        " 1 1.1 1.2 1.3 -1.1  0.2  0.3 10",
        " 2 1.4 1.5 1.6  0.1 -1.2  0.3 20",
        " 3 1.7 1.8 1.9  0.1  0.2 -1.3 30",
        "2 1 2 3",
        " 4 2.1 2.2 2.3 -1.1  0.2  0.3 40",
        " 5 2.4 2.5 2.6  0.1 -1.2  0.3 50",
        " 6 2.7 2.8 2.9  0.1  0.2 -1.3 60",
        "3 1 3 3",
        " 7 3.1 3.2 3.3 -1.1  0.2  0.3 70",
        " 8 3.4 3.5 3.6  0.1 -1.2  0.3 80",
        " 9 3.7 3.8 3.9  0.1  0.2 -1.3 90",
    ))
)

EXAMPLE_DETX_MIXED_IDS = StringIO(
    "\n".join((
        "1 3",
        "8 1 1 3",
        " 83 1.1 1.2 1.3 -1.1  0.2  0.3 10",
        " 81 1.4 1.5 1.6  0.1 -1.2  0.3 20",
        " 82 1.7 1.8 1.9  0.1  0.2 -1.3 30",
        "7 1 2 3",
        " 71 2.1 2.2 2.3 -1.1  0.2  0.3 40",
        " 73 2.4 2.5 2.6  0.1 -1.2  0.3 50",
        " 72 2.7 2.8 2.9  0.1  0.2 -1.3 60",
        "6 1 3 3",
        " 62 3.1 3.2 3.3 -1.1  0.2  0.3 70",
        " 63 3.4 3.5 3.6  0.1 -1.2  0.3 80",
        " 61 3.7 3.8 3.9  0.1  0.2 -1.3 90",
    ))
)

EXAMPLE_DETX_RADIAL = StringIO(
    "\n".join((
        "1 3",
        "1 1 1 4",
        " 1 1 0 0 1 0 0 10",
        " 2 0 1 0 0 1 0 20",
        " 3 -1 0 0 -1 0 0 30",
        " 4 0 -1 0 0 -1 0 40",
        "2 1 2 2",
        " 5 0 0 1 0 0 1 50",
        " 6 0 0 -1 0 0 -1 60",
        "3 1 3 2",
        " 7 1 2 3 1 2 3 70",
        " 8 -3 -2 -1 -3 -2 -1 80",
        "4 2 1 2",
        " 9 0 0 1 0 0 1 90",
        " 10 0 0 -1 0 0 -1 100",
    ))
)


class TestDetector(TestCase):
    def setUp(self):
        self.det = Detector()
        self.det._det_file = EXAMPLE_DETX

    def test_parse_header_extracts_correct_det_id(self):
        self.det._parse_header()
        self.assertEqual(1, self.det.det_id)

    def test_parse_header_extracts_correct_n_doms(self):
        self.det._parse_header()
        self.assertEqual(3, self.det.n_doms)

    def test_parse_doms_maps_each_dom_correctly(self):
        self.det._parse_doms()
        expected = {1: (1, 1, 3), 2: (1, 2, 3), 3: (1, 3, 3)}
        self.assertDictEqual(expected, self.det.doms)

    def test_dom_ids(self):
        self.det._parse_doms()
        self.assertEqual((1, 2, 3), tuple(self.det.dom_ids))

    def test_parse_reset_cache(self):
        self.det._parse_doms()
        assert not self.det._dom_positions
        assert not self.det._pmt_angles
        assert not self.det._xy_positions
        self.det.dom_positions
        self.det.pmt_angles
        self.det.xy_positions
        assert self.det._dom_positions
        assert len(self.det._pmt_angles) == 3
        assert len(self.det._xy_positions) == 1
        self.det.reset_caches()
        assert not self.det._dom_positions
        assert not self.det._pmt_angles
        assert not self.det._xy_positions

    def test_parse_doms_maps_each_dom_correctly_for_mixed_pmt_ids(self):
        self.det._det_file = EXAMPLE_DETX_MIXED_IDS
        self.det._parse_doms()
        expected = {8: (1, 1, 3), 7: (1, 2, 3), 6: (1, 3, 3)}
        self.assertDictEqual(expected, self.det.doms)

    def test_dom_positions(self):
        self.det._parse_doms()
        assert np.allclose([1.49992331, 1.51893187, 1.44185513],
                           self.det.dom_positions[1])
        assert np.allclose([2.49992331, 2.51893187, 2.44185513],
                           self.det.dom_positions[2])
        assert np.allclose([3.49992331, 3.51893187, 3.44185513],
                           self.det.dom_positions[3])

    def test_xy_positions(self):
        self.det._parse_doms()
        assert len(self.det.xy_positions) == 1
        assert np.allclose([1.49992331, 1.51893187], self.det.xy_positions[0])

    def test_correct_number_of_pmts(self):
        self.det._parse_doms()
        assert 9 == len(self.det.pmts)

    def test_pmt_attributes(self):
        self.det._parse_doms()
        assert (1, 2, 3, 4, 5, 6, 7, 8, 9) == tuple(self.det.pmts.pmt_id)
        assert np.allclose([1.1, 1.4, 1.7, 2.1, 2.4, 2.7, 3.1, 3.4, 3.7],
                           self.det.pmts.pos_x)
        assert np.allclose((1.7, 1.8, 1.9), self.det.pmts.pos[2])
        assert np.allclose((0.1, 0.2, -1.3), self.det.pmts.dir[8])

    def test_pmt_index_by_omkey(self):
        self.det._parse_doms()
        assert 5 == self.det._pmt_index_by_omkey[(1, 2, 2)]
        assert 0 == self.det._pmt_index_by_omkey[(1, 1, 0)]
        assert 4 == self.det._pmt_index_by_omkey[(1, 2, 1)]
        assert 1 == self.det._pmt_index_by_omkey[(1, 1, 1)]

    def test_pmt_index_by_pmt_id(self):
        self.det._parse_doms()
        assert 0 == self.det._pmt_index_by_pmt_id[1]

    def test_pmt_with_id_returns_correct_omkeys(self):
        self.det._parse_doms()
        pmt = self.det.pmt_with_id(1)
        assert (1, 1, 0) == (pmt.du, pmt.floor, pmt.channel_id)
        pmt = self.det.pmt_with_id(5)
        assert (1, 2, 1) == (pmt.du, pmt.floor, pmt.channel_id)

    def test_pmt_with_id_returns_correct_omkeys_with_mixed_pmt_ids(self):
        self.det._det_file = EXAMPLE_DETX_MIXED_IDS
        self.det._parse_doms()
        pmt = self.det.pmt_with_id(73)
        assert (1, 2, 1) == (pmt.du, pmt.floor, pmt.channel_id)
        pmt = self.det.pmt_with_id(81)
        assert (1, 1, 1) == (pmt.du, pmt.floor, pmt.channel_id)

    def test_pmt_with_id_raises_exception_for_invalid_id(self):
        self.det._parse_doms()
        with self.assertRaises(KeyError):
            self.det.pmt_with_id(100)

    def test_get_pmt(self):
        self.det._det_file = EXAMPLE_DETX_MIXED_IDS
        self.det._parse_doms()
        pmt = self.det.get_pmt(7, 2)
        assert (1, 2, 2) == (pmt.du, pmt.floor, pmt.channel_id)

    def test_xy_pos(self):
        self.det._parse_doms()
        xy = self.det.xy_positions
        assert xy is not None

    def test_ascii(self):
        detx_string = "\n".join((
            "1 3",
            "1 1 1 3",
            " 1 1.1 1.2 1.3 1.1 2.1 3.1 10.0",
            " 2 1.4 1.5 1.6 4.1 5.1 6.1 20.0",
            " 3 1.7 1.8 1.9 7.1 8.1 9.1 30.0",
            "2 1 2 3",
            " 4 2.1 2.2 2.3 1.2 2.2 3.2 40.0",
            " 5 2.4 2.5 2.6 4.2 5.2 6.2 50.0",
            " 6 2.7 2.8 2.9 7.2 8.2 9.2 60.0",
            "3 1 3 3",
            " 7 3.1 3.2 3.3 1.3 2.3 3.3 70.0",
            " 8 3.4 3.5 3.6 4.3 5.3 6.3 80.0",
            " 9 3.7 3.8 3.9 7.3 8.3 9.3 90.0\n",
        ))
        detx_fob = StringIO(detx_string)

        self.det = Detector()
        self.det._det_file = detx_fob
        self.det._parse_header()
        self.det._parse_doms()
        assert detx_string == self.det.ascii

    def test_ascii_with_mixed_dom_ids(self):
        detx_string = "\n".join((
            "1 3",
            "8 1 1 3",
            " 1 1.1 1.2 1.3 1.1 2.1 3.1 10.0",
            " 2 1.4 1.5 1.6 4.1 5.1 6.1 20.0",
            " 3 1.7 1.8 1.9 7.1 8.1 9.1 30.0",
            "4 1 2 3",
            " 4 2.1 2.2 2.3 1.2 2.2 3.2 40.0",
            " 5 2.4 2.5 2.6 4.2 5.2 6.2 50.0",
            " 6 2.7 2.8 2.9 7.2 8.2 9.2 60.0",
            "9 1 3 3",
            " 7 3.1 3.2 3.3 1.3 2.3 3.3 70.0",
            " 8 3.4 3.5 3.6 4.3 5.3 6.3 80.0",
            " 9 3.7 3.8 3.9 7.3 8.3 9.3 90.0\n",
        ))
        detx_fobj = StringIO(detx_string)

        self.det = Detector()
        self.det._det_file = detx_fobj
        self.det._parse_header()
        self.det._parse_doms()
        assert detx_string == self.det.ascii

    def test_init_from_string(self):
        detx_string = "\n".join((
            "1 3",
            "8 1 1 3",
            " 1 1.1 1.2 1.3 1.1 2.1 3.1 10.0",
            " 2 1.4 1.5 1.6 4.1 5.1 6.1 20.0",
            " 3 1.7 1.8 1.9 7.1 8.1 9.1 30.0",
            "4 1 2 3",
            " 4 2.1 2.2 2.3 1.2 2.2 3.2 40.0",
            " 5 2.4 2.5 2.6 4.2 5.2 6.2 50.0",
            " 6 2.7 2.8 2.9 7.2 8.2 9.2 60.0",
            "9 1 3 3",
            " 7 3.1 3.2 3.3 1.3 2.3 3.3 70.0",
            " 8 3.4 3.5 3.6 4.3 5.3 6.3 80.0",
            " 9 3.7 3.8 3.9 7.3 8.3 9.3 90.0\n",
        ))
        det = Detector(string=detx_string)
        assert 1 == det.n_dus
        assert 3 == det.n_doms

    def test_detx_format_version_1(self):
        det = Detector(filename=data_path('detx/detx_v1.detx'))
        assert 2 == det.n_dus
        assert 6 == det.n_doms
        assert 3 == det.n_pmts_per_dom
        assert 'v1' == det.version
        self.assertListEqual([1.1, 1.2, 1.3], list(det.pmts.pos[0]))
        self.assertListEqual([3.4, 3.5, 3.6], list(det.pmts.pos[7]))
        self.assertListEqual([23.4, 23.5, 23.6], list(det.pmts.pos[16]))

    def test_detx_v1_is_the_same_ascii(self):
        det = Detector(filename=data_path('detx/detx_v1.detx'))
        with open(data_path('detx/detx_v1.detx'), 'r') as fobj:
            assert fobj.read() == det.ascii

    def test_detx_format_version_2(self):
        det = Detector(filename=data_path('detx/detx_v2.detx'))
        assert 2 == det.n_dus
        assert 6 == det.n_doms
        assert 3 == det.n_pmts_per_dom
        assert 256500.0 == det.utm_info.easting
        assert 4743000.0 == det.utm_info.northing
        assert 'WGS84' == det.utm_info.ellipsoid
        assert '32N' == det.utm_info.grid
        assert -2425.0 == det.utm_info.z
        assert 1500000000.1 == det.valid_from
        assert 9999999999.0 == det.valid_until
        assert 'v2' == det.version
        self.assertListEqual([1.1, 1.2, 1.3], list(det.pmts.pos[0]))
        self.assertListEqual([3.4, 3.5, 3.6], list(det.pmts.pos[7]))
        self.assertListEqual([23.4, 23.5, 23.6], list(det.pmts.pos[16]))

    def test_detx_v2_is_the_same_ascii(self):
        det = Detector(filename=data_path('detx/detx_v2.detx'))
        with open(data_path('detx/detx_v2.detx'), 'r') as fobj:
            assert fobj.read() == det.ascii

    def test_detx_format_version_3(self):
        det = Detector(filename=data_path('detx/detx_v3.detx'))
        assert 2 == det.n_dus
        assert 6 == det.n_doms
        assert 3 == det.n_pmts_per_dom
        assert 256500.0 == det.utm_info.easting
        assert 4743000.0 == det.utm_info.northing
        assert 'WGS84' == det.utm_info.ellipsoid
        assert '32N' == det.utm_info.grid
        assert -2425.0 == det.utm_info.z
        assert 1500000000.1 == det.valid_from
        assert 9999999999.0 == det.valid_until
        assert 'v3' == det.version
        self.assertListEqual([1.1, 1.2, 1.3], list(det.pmts.pos[0]))
        self.assertListEqual([3.4, 3.5, 3.6], list(det.pmts.pos[7]))
        self.assertListEqual([23.4, 23.5, 23.6], list(det.pmts.pos[16]))

    def test_detector_repr(self):
        det = Detector(filename=data_path('detx/detx_v3.detx'))
        assert "Detector id: '23', n_doms: 6, dus: [1, 2]" == repr(det)

    def test_detx_format_version_3_with_whitespace(self):
        det = Detector(filename=data_path('detx/detx_v3_whitespace.detx'))
        assert 2 == det.n_dus
        assert 6 == det.n_doms
        assert 3 == det.n_pmts_per_dom
        assert 256500.0 == det.utm_info.easting
        assert 4743000.0 == det.utm_info.northing
        assert 'WGS84' == det.utm_info.ellipsoid
        assert '32N' == det.utm_info.grid
        assert -2425.0 == det.utm_info.z
        assert 1500000000.1 == det.valid_from
        assert 9999999999.0 == det.valid_until
        assert 'v3' == det.version
        self.assertListEqual([1.1, 1.2, 1.3], list(det.pmts.pos[0]))
        self.assertListEqual([3.4, 3.5, 3.6], list(det.pmts.pos[7]))
        self.assertListEqual([23.4, 23.5, 23.6], list(det.pmts.pos[16]))

    def test_detx_format_comments(self):
        det = Detector(filename=data_path('detx/detx_v1.detx'))
        assert len(det.comments) == 0

        det = Detector(filename=data_path('detx/detx_v2.detx'))
        assert len(det.comments) == 0

        det = Detector(filename=data_path('detx/detx_v3.detx'))
        assert len(det.comments) == 2
        assert " a comment line" == det.comments[0]
        assert " another comment line starting with '#'" == det.comments[1]

    def test_comments_are_written(self):
        det = Detector(filename=data_path('detx/detx_v3.detx'))
        det.add_comment("foo")
        assert 3 == len(det.comments)
        assert det.comments[2] == "foo"
        assert "# foo" == det.ascii.splitlines()[2]

    def test_detx_v3_is_the_same_ascii(self):
        det = Detector(filename=data_path('detx/detx_v3.detx'))
        with open(data_path('detx/detx_v3.detx'), 'r') as fobj:
            assert fobj.read() == det.ascii

    def test_translate_detector(self):
        self.det._parse_doms()
        t_vec = np.array([1, 2, 3])
        orig_pos = self.det.pmts.pos.copy()
        self.det.translate_detector(t_vec)
        assert np.allclose(orig_pos + t_vec, self.det.pmts.pos)

    def test_translate_detector_updates_xy_positions(self):
        self.det._parse_doms()
        t_vec = np.array([1, 2, 3])
        orig_xy_pos = self.det.xy_positions.copy()
        self.det.translate_detector(t_vec)
        assert np.allclose(orig_xy_pos + t_vec[:2], self.det.xy_positions)

    def test_translate_detector_updates_dom_positions(self):
        self.det._parse_doms()
        t_vec = np.array([1, 2, 3])
        orig_dom_pos = deepcopy(self.det.dom_positions)
        self.det.translate_detector(t_vec)
        for dom_id, pos in self.det.dom_positions.items():
            assert np.allclose(orig_dom_pos[dom_id] + t_vec, pos)

    def test_rotate_dom_by_yaw(self):
        det = Detector()
        det._det_file = EXAMPLE_DETX_RADIAL
        det._parse_doms()
        # here, only one PMT is checked
        dom_id = 1
        heading = 23
        channel_id = 0
        pmt_dir = det.pmts[det.pmts.dom_id == dom_id].dir[channel_id].copy()
        pmt_dir_rot = qrot_yaw(pmt_dir, heading)
        det.rotate_dom_by_yaw(dom_id, heading)
        assert np.allclose(
            pmt_dir_rot, det.pmts[det.pmts.dom_id == dom_id].dir[channel_id]
        )
        assert np.allclose([0.92050485, 0.39073113, 0],
                           det.pmts[det.pmts.dom_id == dom_id].pos[channel_id])

    def test_rotate_dom_set_by_step_by_360_degrees(self):
        det = Detector()
        det._det_file = EXAMPLE_DETX_RADIAL
        det._parse_doms()
        dom_id = 1
        channel_id = 0
        pmt_dir = det.pmts[det.pmts.dom_id == dom_id].dir[channel_id].copy()
        pmt_pos = det.pmts[det.pmts.dom_id == dom_id].pos[channel_id].copy()
        for i in range(36):
            det.rotate_dom_by_yaw(dom_id, 10)
        pmt_dir_rot = det.pmts[det.pmts.dom_id == dom_id].dir[channel_id]
        assert np.allclose(pmt_dir, pmt_dir_rot)
        pmt_pos_rot = det.pmts[det.pmts.dom_id == dom_id].pos[channel_id]
        assert np.allclose(pmt_pos, pmt_pos_rot)

    def test_rotate_du_by_yaw_step_by_step_360_degrees(self):
        det = Detector()
        det._det_file = EXAMPLE_DETX_RADIAL
        det._parse_doms()
        du = 2
        pmt_dir = det.pmts[det.pmts.du == du].dir.copy()
        pmt_pos = det.pmts[det.pmts.du == du].pos.copy()
        pmt_dir_other_dus = det.pmts[det.pmts.du != du].dir.copy()
        pmt_pos_other_dus = det.pmts[det.pmts.du != du].pos.copy()
        for i in range(36):
            det.rotate_du_by_yaw(du, 10)
        pmt_dir_rot = det.pmts[det.pmts.du == du].dir
        pmt_pos_rot = det.pmts[det.pmts.du == du].pos
        assert np.allclose(pmt_dir, pmt_dir_rot)
        assert np.allclose(pmt_pos, pmt_pos_rot)
        assert np.allclose(pmt_dir_other_dus, det.pmts[det.pmts.du != du].dir)
        assert np.allclose(pmt_pos_other_dus, det.pmts[det.pmts.du != du].pos)

    def test_rescale_detector(self):
        self.det._parse_doms()
        dom_positions = deepcopy(self.det.dom_positions)
        scale_factor = 2
        self.det.rescale(scale_factor)
        for dom_id, dom_pos in self.det.dom_positions.items():
            assert np.allclose(dom_pos, dom_positions[dom_id] * scale_factor)

    def test_dom_table(self):
        self.det._parse_doms()
        dt = self.det.dom_table
        assert 3 == len(dt)
        assert np.allclose([1, 2, 3], dt.dom_id)
        assert np.allclose([1, 1, 1], dt.du)
        assert np.allclose([1, 2, 3], dt.floor)
        assert np.allclose([1.49992331, 2.49992331, 3.49992331], dt.pos_x)
        assert np.allclose([1.51893187, 2.51893187, 3.51893187], dt.pos_y)
        assert np.allclose([1.44185513, 2.44185513, 3.44185513], dt.pos_z)

    def test_dom_table_with_another_detx(self):
        det = Detector()
        det._det_file = EXAMPLE_DETX_RADIAL
        det._parse_doms()

        dt = det.dom_table
        assert 4 == len(dt)
        assert np.allclose([1, 2, 3, 4], dt.dom_id)
        assert np.allclose([1, 1, 1, 2], dt.du)
        assert np.allclose([1, 2, 3, 1], dt.floor)
        assert np.allclose([0, 0, 0, 0], dt.pos_x)
        assert np.allclose([0, 0, 0, 0], dt.pos_y)
        assert np.allclose([0, 0, 0, 0], dt.pos_z)

    def test_center_of_mass(self):
        det = Detector()
        det._det_file = EXAMPLE_DETX
        det._parse_doms()

        assert np.allclose([2.4, 2.5, 2.6], det.com)

    def test_center_of_mass_with_another_detx(self):
        det = Detector()
        det._det_file = EXAMPLE_DETX_RADIAL
        det._parse_doms()

        assert np.allclose([-0.2, 0., 0.2], det.com)


    def test_jdetectordb_output_with_detx_v3(self):
        det = Detector(data_path("detx/D_ORCA006_t.A02181836.p.A02181837.r.A02182001.detx"))
        assert det.utm_info is not None
        assert det.utm_info.ellipsoid == "WGS84"
        assert det.utm_info.grid == "32N"
        assert det.utm_info.easting == 256500.0
        assert det.utm_info.northing == 4743000.0
        assert det.utm_info.z == -2440.0



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
Calibration

"""

from km3pipe import Pipeline, Module
from km3pipe.calib import Calibration
from km3pipe.io import EvtPump
from km3modules.common import StatusBar
import os
import time

PATH = '/Users/tamasgal/Data/KM3NeT'
DETX = 'Detector/km3net_jul13_90m_r1494_corrected.detx'
DATA = 'km3net_jul13_90m_muatm10T23.km3_v5r1.JTE.evt'


class PrintPositions(Module):
    def process(self, blob):
        print("Hit positions:")
        print(blob['Hits'].pos)
        time.sleep(0.1)
        return blob

    def finish(self):
        time.sleep(2)


pipe = Pipeline()
pipe.attach(EvtPump, filename=os.path.join(PATH, DATA))
pipe.attach(StatusBar)
pipe.attach(
    Calibration, timeit=True, apply=True, filename=os.path.join(PATH, DETX)
)
pipe.attach(PrintPositions, timeit=True)
pipe.drain(3)



#!/usr/bin/env python
"""
Convert jupyter notebook to sphinx gallery notebook styled examples.
Needs pypandoc (`pip install pypandoc`).

Usage:
    nb2sphx NOTEBOOK

"""

import json
import pypandoc as pdoc

from km3pipe import version


def convert_ipynb_to_gallery(file_name):
    """
    Blatantly stolen + adapted from
    https://gist.github.com/wuhuikai/4a7ceb8bc52454e17a4eb8327d538d85

    """
    python_file = ""

    nb_dict = json.load(open(file_name))
    cells = nb_dict['cells']

    for i, cell in enumerate(cells):
        if i == 0:
            assert cell['cell_type'] == 'markdown', \
                'First cell has to be markdown'

            md_source = ''.join(cell['source'])
            rst_source = pdoc.convert_text(md_source, 'rst', 'md')
            python_file = '"""\n' + rst_source + '\n"""'
        else:
            if cell['cell_type'] == 'markdown':
                md_source = ''.join(cell['source'])
                rst_source = pdoc.convert_text(md_source, 'rst', 'md')
                commented_source = '\n'.join([
                    '# ' + x for x in rst_source.split('\n')
                ])
                python_file = python_file + '\n\n\n' + '#' * 70 + '\n' + \
                    commented_source
            elif cell['cell_type'] == 'code':
                source = ''.join(cell['source'])
                python_file = python_file + '\n' * 2 + source

    open(file_name.replace('.ipynb', '.py'), 'w').write(python_file)


def main():
    from docopt import docopt
    args = docopt(__doc__, version=version)
    fname = args['NOTEBOOK']
    convert_ipynb_to_gallery(fname)


if __name__ == '__main__':
    main()



#!/usr/bin/env python
# coding=utf-8
# Filename: qrunqaqc.py
# Author: Tamas Gal <tgal@km3net.de>
"""
Submits jobs which analyse a batch of runs with ``JQAQC.sh`` to retrieve the
quality parameters. The obtained data is then optionally uploaded to the
runsummarynumbers table of the KM3NeT database.

The submitted run IDs are logged in ``BLACKLIST_{DET_ID}.txt`` and skipped
upon subsequential runs. The temporary quality parameters are stored in
the folder ``qparams`` and the job logs can be found in ``qlogs``.

To start from scratch, delete the above mentioned files. Runs which are were
already processed and present in the runsummarynumbers table are always
skipped.

Usage:
    qrunqaqc [options] DET_ID
    qrunqaqc (-h | --help)

Options:
    -b BATCH_SIZE  Number of runs to process in a single job.
    -j MAX_JOBS    Maximum number of jobs to submit [default: 100].
    -u             Upload summary data to the database automatically.
    -q             Dryrun, don't submit the jobs.
    -h --help      Show this screen.

"""
from collections import defaultdict
import math
import os
import subprocess
import time

from tqdm import tqdm
import km3pipe as kp

kp.logger.get_logger('km3pipe.db').setLevel('CRITICAL')

ESTIMATED_TIME_PER_RUN = 60 * 10    # [s]


class QAQCAnalyser(object):
    """Determines  run quality parameters and uploads them to the DB"""
    def __init__(self, det_id, should_upload_to_db, log_file="qrunqaqc.log"):
        self.det_id = det_id
        self.should_upload_to_db = should_upload_to_db

        self.log = kp.logger.get_logger("qrunqaqc", filename=log_file)

        self.log.info("QAQC analysis started for detector ID %s", det_id)
        if should_upload_to_db:
            self.log.info(
                "The acquired data will be automatically updloaded to "
                "the database after each successful run processing."
            )

        self.jpp_version = kp.tools.get_jpp_version()
        if self.jpp_version is None:
            self.log.critical("Please load a Jpp environment")
            raise SystemExit()
        else:
            print(
                "Run quality determination using Jpp '{}'".format(
                    self.jpp_version
                )
            )

        self.sds = kp.db.StreamDS()
        self.db = kp.db.DBManager()
        self.det_oid = self.db.get_det_oid(det_id)

        self.runtable = self.sds.get("runs", detid=self.det_id)

        self.available_run_files = []

        cwd = os.getcwd()
        self.outdir = os.path.join(cwd, "qparams")
        self.jobdir = os.path.join(cwd, "qjobs")
        for path in [self.outdir, self.jobdir]:
            if not os.path.exists(path):
                os.makedirs(path)

        self.blacklist = os.path.join(
            cwd, "blacklist_{}.txt".format(self.det_id)
        )

        self.stats = defaultdict(int)

        self._qparams = None
        self._columns = None
        self._blacklisted_run_ids = None

    def add_to_blacklist(self, run_id):
        """Put the run_id into a det_id specific blacklist"""
        with open(self.blacklist, "a") as fobj:
            fobj.write("{}\n".format(run_id))

    @property
    def blacklisted_run_ids(self):
        """The blacklisted run_ids"""
        if self._blacklisted_run_ids is None:
            if os.path.exists(self.blacklist):
                with open(self.blacklist) as fobj:
                    self._blacklisted_run_ids = set(
                        int(l) for l in fobj.readlines()
                    )
            else:
                self._blacklisted_run_ids = set()
        return self._blacklisted_run_ids

    def retrieve_available_runs(self):
        files = kp.tools.ifiles(
            "data/raw/sea/KM3NeT_{:08d}".format(self.det_id)
        )
        self.available_run_files = {f.path: f.size for f in files}

    def run(self, batch_size, max_jobs, dryrun):
        """Walk through the runtable and submit batches of runs"""

        run_ids = self.runtable.RUN.values
        print("{} runs in total".format(len(run_ids)))

        try:
            already_processed_runs_ids = self.sds.get(
                "runsummarynumbers",
                detid=self.det_oid,
                minrun=min(run_ids),
                maxrun=max(run_ids),
                source_name=self.jpp_version,
                parameter_name="livetime_s"    # to lower the request size
            ).RUN.values
        except AttributeError:
            already_processed_runs_ids = set()
        else:
            print(
                "{} runs are already processed and available in the DB".format(
                    len(already_processed_runs_ids)
                )
            )

        if self.blacklisted_run_ids:
            print(
                "Skipping {} runs since they were already submitted "
                "and may be in the job queue. Delete the blacklist file "
                "to resubmit them next time this script is ran.".format(
                    len(self.blacklisted_run_ids)
                )
            )

        n_jobs = 0
        run_ids_to_process = []
        run_ids_to_check = sorted(
            set(run_ids) - set(already_processed_runs_ids) -
            self.blacklisted_run_ids,
            reverse=True
        )

        print("Checking runs, retrieving file list from iRODS...")

        self.retrieve_available_runs()

        for run_id in tqdm(run_ids_to_check):
            if n_jobs >= max_jobs:
                self.log.warning(
                    "Maximum number of jobs reached (%s), "
                    "proceeding with batch submission", max_jobs
                )
                break
            self.log.info("Checking run '{}'".format(run_id))

            irods_path = kp.tools.irods_path(self.det_id, run_id)
            if irods_path in self.available_run_files.keys():
                run_ids_to_process.append(run_id)
                if batch_size and len(run_ids_to_process) % batch_size == 0:
                    n_jobs += 1
                continue
            else:
                self.log.info(
                    "  -> no file found on iRODS or an iRODS error "
                    "occured for run {}".format(run_id)
                )
                self.stats['Missing data or iRODS error'] += 1

        total_runs_to_process = len(run_ids_to_process)

        if total_runs_to_process == 0:
            print("No runs to process.")
        else:

            if batch_size is None:
                batch_size = int(
                    math.ceil(total_runs_to_process / float(max_jobs))
                )

            print(
                "Proceeding with {} runs distributed over {} jobs, {} runs/job"
                .format(total_runs_to_process, max_jobs, batch_size)
            )

            run_id_chunks = kp.tools.chunks(run_ids_to_process, batch_size)

            self.pbar_runs = tqdm(
                total=len(run_ids_to_process),
                desc="Submitting runs",
                unit='run'
            )

            for run_ids in tqdm(run_id_chunks, desc="Jobs"):

                self.submit_batch(run_ids, dryrun=dryrun)

            self.pbar_runs.close()

        self.print_stats()

    def print_stats(self):
        """Print a summary of the collected statistics"""
        print("\n")
        print("Summary")
        print("=======")
        for key, value in self.stats.items():
            print("  {}: {}".format(key, value))

    def submit_batch(self, run_ids, dryrun):
        """Submit a QAQC.sh job for a given list of run IDs"""

        filesizes = []

        s = kp.shell.Script()
        for run_id in run_ids:
            self.log.info("adding run %s (det_id %s", run_id, self.det_id)
            xrootd_path = kp.tools.xrootd_path(self.det_id, run_id)
            self.log.info("xrootd path: %s", xrootd_path)
            size = self.available_run_files[kp.tools.irods_path(
                self.det_id, run_id
            )]
            self.log.info("filesize: %s (bytes)", size)
            filesizes.append(size)

            root_filename = os.path.basename(xrootd_path)

            out_filename = os.path.join(
                self.outdir, "{}_{}_qparams.csv".format(self.det_id, run_id)
            )
            t0set = self.runtable[self.runtable.RUN == run_id
                                  ].T0_CALIBSETID.values[0]

            self.log.info("adding run '%s' with t0set '%s'", run_id, t0set)

            s.separator()
            s.echo("Processing run {}".format(run_id))
            s.separator('-')
            s.add("km3pipe detx {} -t {} -o d.detx".format(self.det_id, t0set))
            s.add("xrdcp {} {}".format(xrootd_path, root_filename))
            s.add("echo '{}'> {}".format(" ".join(self.columns), out_filename))
            s.add(
                "JQAQC.sh d.detx {} "
                "| awk '{{$1=$1;print}}' | tr '\\n' ' ' >> {}".format(
                    root_filename, out_filename
                )
            )
            if self.should_upload_to_db:
                s.add("streamds upload {}".format(out_filename))
            s.add("rm -f {}".format(root_filename))

            self.add_to_blacklist(run_id)
            self.pbar_runs.update(1)
            self.stats['Number of submitted runs'] += 1

        walltime = time.strftime(
            '%H:%M:%S', time.gmtime(ESTIMATED_TIME_PER_RUN * len(run_ids))
        )

        fsize = int(max(filesizes) / 1024 / 1024 * 1.1 + 100)

        identifier = "QAQC_{}_{}-{}".format(
            self.det_id, run_ids[0], run_ids[-1]
        )

        job_script = kp.shell.qsub(
            s,
            identifier,
            vmem='4G',
            fsize='{}M'.format(fsize),
            xrootd=True,
            walltime=walltime,
            silent=True,
            dryrun=dryrun
        )
        self.log.info("  => job with %s runs submitted", len(run_ids))

        jobfile = os.path.join(self.jobdir, identifier + ".sh")
        with open(jobfile, 'w') as fobj:
            fobj.write(job_script)
        self.log.info("     job file have been saved to {}".format(jobfile))
        if dryrun:
            self.stats['Number of dryrun jobs'] += 1
        else:
            self.stats['Number of submitted jobs'] += 1

    def retrieve_qparams(self):
        """Returns a list of quality parameters determined by JQAQC.sh"""
        command = "JQAQC.sh -h"
        try:
            qparams = subprocess.getoutput(command)
        except AttributeError:
            qparams = subprocess.check_output(
                command.split(), stderr=subprocess.STDOUT
            )
        return qparams.split('\n')[1].split()

    @property
    def columns(self):
        """Get the ordered columns for the quality parameter output"""
        if self._columns is None:
            qparams = self.retrieve_qparams()
            # Adapting to runsummarynumbers naming convention
            # Warning: for retrieving data, it's called "source_name"
            qparams[qparams.index('GIT')] = "source"
            qparams[qparams.index('detector')] = "det_id"

            self._columns = qparams
        return self._columns


def main():
    from docopt import docopt
    args = docopt(__doc__)

    try:
        batch_size = int(args['-b'])
    except TypeError:
        batch_size = None

    qaqc = QAQCAnalyser(
        det_id=int(args['DET_ID']), should_upload_to_db=args['-u']
    )
    qaqc.run(
        batch_size=batch_size,
        max_jobs=int(args['-j']),
        dryrun=bool(args['-q'])
    )


if __name__ == '__main__':
    main()



#!/usr/bin/env python
# vim: ts=4 sw=4 et

from collections import deque, defaultdict
from functools import partial
from io import BytesIO
import os

import km3pipe as kp
from km3modules.plot import plot_dom_parameters
import km3pipe.style

km3pipe.style.use('km3pipe')

PLOTS_PATH = 'km3web/plots'
cal = kp.calib.Calibration(det_id=29)
detector = cal.detector


class DOMActivityPlotter(kp.Module):
    def configure(self):
        self.index = 0
        self.last_activity = defaultdict(partial(deque, maxlen=4000))
        self.cuckoo = kp.time.Cuckoo(60, self.create_plot)

    def process(self, blob):
        self.index += 1
        if self.index % 30:
            return blob

        tag = str(blob['CHPrefix'].tag)

        if not tag == 'IO_SUM':
            return blob

        data = blob['CHData']
        data_io = BytesIO(data)
        preamble = kp.io.daq.DAQPreamble(file_obj=data_io)    # noqa
        summaryslice = kp.io.daq.DAQSummaryslice(file_obj=data_io)
        timestamp = summaryslice.header.time_stamp

        for dom_id, _ in summaryslice.summary_frames.items():
            du, dom, _ = detector.doms[dom_id]
            self.last_activity[(du, dom)] = timestamp

        self.cuckoo.msg()

        return blob

    def create_plot(self):
        print(self.__class__.__name__ + ": updating plot.")
        filename = os.path.join(PLOTS_PATH, 'dom_activity.png')
        now = kp.time.tai_timestamp()
        delta_ts = {}
        for key, timestamp in self.last_activity.items():
            delta_ts[key] = now - timestamp
        plot_dom_parameters(
            delta_ts,
            detector,
            filename,
            'last activity [s]',
            "DOM Activity - via Summary Slices",
            vmin=0.0,
            vmax=15 * 60
        )


pipe = kp.Pipeline()
pipe.attach(
    kp.io.ch.CHPump,
    host='127.0.0.1',
    port=5553,
    tags='IO_SUM',
    timeout=60 * 60 * 24 * 7,
    max_queue=2000
)
pipe.attach(kp.io.daq.DAQProcessor)
pipe.attach(DOMActivityPlotter)
pipe.drain()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# vim: ts=4 sw=4 et
"""
========
HDF5Sink
========

This examples show how to create dummy (event) data, dump them to
an HDF5 file using the ``kp.io.HDF5Sink`` module and how to read them back
with pandas or the ``kp.io.HDF5Pump`` module.

"""

import km3pipe as kp
import pandas as pd
import numpy as np

filename = "reco_results.h5"

####################################################################
# The ``DataGenerator`` is a dumm pump which creates random event
# data and a common header.


class DataGenerator(kp.Module):
    def configure(self):
        self.header = kp.Table({'is_mc': True},
                               h5singleton=True,
                               h5loc='/header')

    def process(self, blob):

        n = np.random.randint(2, 20)
        tab = kp.Table({
            'lambda': np.random.rand(n),
            'energy': np.random.rand(n) * 1000
        },
                       h5loc='/reco',
                       name='Quality Parameter')
        event_info = kp.Table({"run_id": np.random.randint(10000, 20000)},
                              h5loc='/event_info')
        blob['Reco'] = tab
        blob['EventInfo'] = event_info
        blob['Header'] = self.header
        return blob


####################################################################
# The ``printer`` just prints the blob, obviously.
# data and a common header.


def printer(blob):
    print(blob)
    return blob


####################################################################
# The following pipeline will do 10 iterations and dump the data
# properly formatted into our output file, ``reso_results.h5``

pipe = kp.Pipeline()
pipe.attach(DataGenerator)
pipe.attach(printer)
pipe.attach(kp.io.HDF5Sink, filename=filename)
pipe.drain(10)

####################################################################
# Pandas offers an easy way to read the generated data. As you can
# see, no km3pipe is needed, the only requirement is ``hdf5lib``
# and of course some kind of wrapper library for easy acces, in this case
# Pandas but you can use h5py, pytables directly or other languages.

header = pd.read_hdf(filename, "/header")
event_info = pd.read_hdf(filename, "/event_info")
reco = pd.read_hdf(filename, "/reco")
print(reco.head(3))

####################################################################
# This last section shows how to read the created data with km3pipe
# pipelines.


def inspector(blob):
    reco = blob['Reco']
    lambdas = reco['lambda']
    print(lambdas, type(lambdas), lambdas.dtype)
    return blob


pipe = kp.Pipeline()
pipe.attach(kp.io.HDF5Pump, filename=filename)
pipe.attach(printer)
pipe.attach(inspector)
pipe.drain()



#!/usr/bin/env python
# vim: ts=4 sw=4 et

from datetime import datetime
import os
from queue import Queue, Empty
import shutil
import threading

import matplotlib
# Force matplotlib to not use any Xwindows backend.
matplotlib.use('Agg')    # noqa
import matplotlib.pyplot as plt
import matplotlib.ticker as ticker
import matplotlib.dates as md
import numpy as np

from km3pipe import Pipeline, Module
from km3pipe.calib import Calibration
from km3pipe.io import CHPump
from km3pipe.io.daq import DAQProcessor
import km3pipe.style

km3pipe.style.use('km3pipe')

PLOTS_PATH = 'www/plots'
N_DOMS = 18
N_DUS = 2
cal = Calibration(det_id=14)
detector = cal.detector

xfmt = md.DateFormatter('%Y-%m-%d %H:%M')
lock = threading.Lock()


class ZTPlot(Module):
    def configure(self):
        self.run = True
        self.max_queue = 3
        self.queue = Queue()
        self.thread = threading.Thread(target=self.plot).start()

    def process(self, blob):
        if 'Hits' not in blob:
            return blob

        hits = blob['Hits'].serialise(to='pandas')
        cal.apply(hits)
        e_info = blob['EventInfo']

        if len(np.unique(hits[hits.triggered == True].du)) < 2:    # noqa
            print("Skipping...")
            return blob

        print("OK")
        # print("Event queue size: {0}".format(self.queue.qsize()))
        if self.queue.qsize() < self.max_queue:
            self.queue.put((e_info, hits))

        return blob

    def plot(self):
        while self.run:
            try:
                e_info, hits = self.queue.get(timeout=50)
            except Empty:
                continue
            with lock:
                self.create_plot(e_info, hits)

    def create_plot(self, e_info, hits):
        print(self.__class__.__name__ + ": updating plot.")
        dus = set(hits.du)

        n_plots = len(dus)
        n_cols = int(np.ceil(np.sqrt(n_plots)))
        n_rows = int(n_plots / n_cols) + (n_plots % n_cols > 0)
        fig, axes = plt.subplots(
            ncols=n_cols,
            nrows=n_rows,
            sharex=True,
            sharey=True,
            figsize=(16, 8)
        )

        for ax, (du, du_hits) in zip(axes.flat, hits.groupby("du")):
            trig_hits = du_hits[du_hits.triggered == True]    # noqa

            ax.scatter(du_hits.time, du_hits.pos_z, c='#09A9DE', label='hit')
            ax.scatter(
                trig_hits.time,
                trig_hits.pos_z,
                c='#FF6363',
                label='triggered hit'
            )
            ax.set_title('DU{0}'.format(du), fontsize=16, fontweight='bold')

        for ax in axes.flat:
            ax.tick_params(labelsize=16)
            ax.yaxis.set_major_locator(ticker.MultipleLocator(200))
            xlabels = ax.get_xticklabels()
            for label in xlabels:
                label.set_rotation(45)

        plt.suptitle(
            "FrameIndex {0}, TriggerCounter {1}\n{2} UTC".format(
                e_info.frame_index, e_info.trigger_counter,
                datetime.utcfromtimestamp(e_info.utc_seconds)
            ),
            fontsize=16
        )
        fig.text(0.5, 0.01, 'time [ns]', ha='center')
        fig.text(0.08, 0.5, 'z [m]', va='center', rotation='vertical')
        #        plt.tight_layout()

        filename = 'ztplot'
        f = os.path.join(PLOTS_PATH, filename + '.png')
        f_tmp = os.path.join(PLOTS_PATH, filename + '_tmp.png')
        plt.savefig(f_tmp, dpi=120, bbox_inches="tight")
        plt.close('all')
        shutil.move(f_tmp, f)

    def finish(self):
        self.run = False
        if self.thread is not None:
            self.thread.stop()


pipe = Pipeline()
pipe.attach(
    CHPump,
    host='127.0.0.1',
    tags='IO_EVT, IO_SUM',
    timeout=60 * 60 * 24 * 7,
    max_queue=2000
)
pipe.attach(DAQProcessor)
pipe.attach(ZTPlot)
pipe.drain()



#!/usr/bin/env python
"""
=================
Slackbot for ELOG
=================

"""
import os
import re

import pyinotify

from pyslack import SlackClient    # pip install pyslack-real

wm = pyinotify.WatchManager()
mask = pyinotify.IN_CLOSE_WRITE

URL = 'http://elog.km3net.de'
# LOGBOOK_PATH = '/home/elog/elog'
LOGBOOK_PATH = '/usr/local/elog/logbooks'
BOTNAME = 'ELOG'
DEFAULT_DESTINATION = '#elog'
DESTINATIONS = {
    'Operations_IT': '#operations_it',
    'Operations_FR': '#operations_fr',
    #    'Qualification': '#elog',
    #    'DOM_Integration': '#elog',
    #    'DU_Integration': '#elog',
    #    'DAQ_Readout': '#elog',
    #    'Electronics': '#elog',
    #    'Analysis': '#elog',
    'Computing and Software': '#software',
}

slack = SlackClient("YOUR_SLACK_API_TOKEN_HERE")


class ElogEntry(object):
    def __init__(self, logbook, msg_id=None):
        self.logbook = logbook
        self.id = msg_id
        self.header = {}
        self.content = ''

    @property
    def author(self):
        return self._lookup('Author')

    @property
    def type(self):
        return self._lookup('Type')

    @property
    def subject(self):
        return self._lookup('Subject')

    @property
    def url(self):
        escaped = re.sub(r"[\s_]+", '+', self.logbook)
        return URL + '/' + escaped + '/' + str(self.id)

    def _lookup(self, key):
        try:
            return self.header[key]
        except KeyError:
            return "Unknown"

    def __repr__(self):
        return (
            "ELOG entry from *{0.author}* in *{0.logbook}* (_{0.type}_):\n"
            "*{0.subject}*\n{1}...\n<{0.url}>".format(
                self, self.content[:150]
            )
        )


class ElogEntryBundle(object):
    def __init__(self, pathname):
        self.logbook = dirname(pathname)

        with open(pathname) as input_file:
            self.entries = self._parse_entries(input_file)

    @property
    def newest_entry(self):
        return self.entries[-1]

    def _parse_entries(self, input_file, msg_id=None):
        if msg_id is None:
            line = input_file.readline()
            if not line.startswith('$@MID@$:'):
                raise ValueError("Not an ELOG-entry file!")
            msg_id = int(line.split(':')[1])

        entry = ElogEntry(self.logbook, msg_id)
        for line in input_file:
            if line.startswith(40 * '='):
                break
            parameter, value = re.findall(r'([^:.]*): (.*)', line)[0]
            entry.header[parameter] = value
        for line in input_file:
            if line.startswith('$@MID@$:'):
                msg_id = int(line.split(':')[1])
                return [entry] + self._parse_entries(input_file, msg_id)
            else:
                entry.content += line
        return [entry]


class EventHandler(pyinotify.ProcessEvent):
    def my_init(self):
        self.logged_ids = []

    def process_IN_CLOSE_WRITE(self, event):
        if not self._is_valid_filetype(event.pathname):
            print("Not a valid filetype: {0}".format(event.pathname))
            return

        try:
            elog_bundle = ElogEntryBundle(event.pathname)
            elog_entry = elog_bundle.newest_entry
        except ValueError:
            print("Could not parse '{0}'. Ignoring...".format(event.pathname))
            return

        try:
            destination = DESTINATIONS[elog_entry.logbook]
        except KeyError:
            print(
                "No destination for logbook '{0}'. Using only default...".
                format(elog_entry.logbook)
            )
            destination = None
        finally:
            if elog_entry.id in self.logged_ids:
                return    # For now skip, since it often duplicates!
                pre = 'Updated '
            else:
                pre = ''
                self.logged_ids.append(elog_entry.id)
            message = pre + str(elog_entry)
            print(42 * "#")
            print("Filename: " + event.name)
            print("Elog Entry ID: " + str(elog_entry.id))
            print(42 * "-")
            print(message)
            print(42 * "/")
            if destination:
                slack.chat_post_message(destination, message, username=BOTNAME)
            # else:
            slack.chat_post_message(
                DEFAULT_DESTINATION, message, username=BOTNAME
            )

    def _is_valid_filetype(self, path):
        return path.endswith('.log')


def dirname(filepath):
    return os.path.basename(os.path.dirname(filepath))


handler = EventHandler()
notifier = pyinotify.Notifier(wm, handler)
wdd = wm.add_watch(LOGBOOK_PATH, mask, rec=True)

notifier.loop()



#!/usr/bin/env python
# Filename: ligiermirror.py
# Author: Tamas Gal <tgal@km3net.de>
# vim: ts=4 sw=4 et
"""
Subscribes to given tag(s) and sends them to another Ligier.

Usage:
    ligiermirror [options] SOURCE_IP
    ligiermirror (-h | --help)

Options:
    -t TARGET_IP    Target IP [default: 127.0.0.1].
    -p PORT         Source port [default: 5553].
    -q PORT         Target port [default: 5553].
    -m TAGS         Comma separated message tags [default: IO_EVT, IO_SUM].
    -s QUEUE        Maximum queue size for messages [default: 20000].
    -x TIMEOUT      Connection timeout in seconds [default: 604800].
    -d DEBUG_LEVEL  Debug level (DEBUG, INFO, WARNING, ...) [default: WARNING].
    -h --help       Show this screen.

"""
import socket

import km3pipe as kp


class LigierSender(kp.Module):
    """Forwards a message to another ligier"""
    def configure(self):
        target_ip = self.get("target_ip", default="127.0.0.1")
        port = self.get("port", default=5553)
        self.socket = socket.socket()
        self.client = self.socket.connect((target_ip, port))

    def process(self, blob):
        self.socket.send(blob["CHPrefix"].data + blob["CHData"])

    def finish(self):
        self.socket.close()


def main():
    """The main script"""
    from docopt import docopt
    args = docopt(__doc__, version=kp.version)

    kp.logger.set_level("km3pipe", args['-d'])

    pipe = kp.Pipeline()
    pipe.attach(
        kp.io.ch.CHPump,
        host=args['SOURCE_IP'],
        port=int(args['-p']),
        tags=args['-m'],
        timeout=int(args['-x']),
        max_queue=int(args['-s']),
        show_statistics=True
    )
    pipe.attach(LigierSender, target_ip=args['-t'], port=int(args['-q']))
    pipe.drain()


if __name__ == '__main__':
    main()



# Filename: parser.py
# -*- coding: utf-8 -*-
# pylint: disable=locally-disabled
"""
A collection of parsers.

"""

from km3pipe import Module
from km3pipe.io.daq import DAQPreamble, DAQSummaryslice, DAQEvent
from io import StringIO


class CHParser(Module):
    """A parser for ControlHost data.

    This parser will choose the correct class to parse the binary data
    for given `tags`.

    """
    parse_map = {
        'IO_SUM': ['DAQSummaryslice', DAQSummaryslice],
        'IO_EVT': ['DAQEvent', DAQEvent],
    }

    def configure(self):
        self.tags = self.get('tags') or []

    def process(self, blob):
        if not ('CHData' in blob and 'CHPrefix' in blob):
            return blob

        tag = str(blob['CHPrefix'].tag)

        if tag not in self.tags and tag not in self.parse_map:
            return blob

        data = blob['CHData']
        data_io = StringIO(data)
        preamble = DAQPreamble(file_obj=data_io)    # noqa

        blob_key, ParserClass = self.parse_map[tag]
        blob[blob_key] = ParserClass(file_obj=data_io)

        return blob



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
======================
Basic Analysis Example
======================

"""

# Authors: Tamás Gál <tgal@km3net.de>, Moritz Lotze <mlotze@km3net.de>
# License: BSD-3
# Date: 2017-10-10
# Status: Under construction...

#####################################################
# Preparation
# -----------
# The very first thing we do is importing our libraries and setting up
# the Jupyter Notebook environment.

import matplotlib.pyplot as plt    # our plotting module
import pandas as pd    # the main HDF5 reader
import numpy as np    # must have
import km3pipe as kp    # some KM3NeT related helper functions
import seaborn as sns    # beautiful statistical plots!

#####################################################
# this is just to make our plots a bit "nicer", you can skip it
import km3pipe.style
km3pipe.style.use("km3pipe")

#####################################################
# Accessing the Data File(s)
# --------------------------
# In the following, we will work with one random simulation file with
# reconstruction information from JGandalf which has been converted
# from ROOT to HDF5 using the ``tohdf5`` command line tool provided by
# ``KM3Pipe``.
#
# You can find the documentation here:
# https://km3py.pages.km3net.de/km3pipe/cmd.html#tohdf

#####################################################
# Note for Lyon Users
# ~~~~~~~~~~~~~~~~~~~
# If you are working on the Lyon cluster, you can activate the latest KM3Pipe
# with the following command (put it in your ``~/.bashrc`` to load it
# automatically in each shell session)::
#
#     source $KM3NET_THRONG_DIR/src/python/pyenv.sh

#####################################################
# Converting from ROOT to HDF5 (if needed)
# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
#
# Choose a file (take e.g. one from /in2p3/km3net/mc/...),
# load the appropriate Jpp/Aanet version and convert it via::
#
#     tohdf5 --aa-format=the_file.root --ignore-hits --skip-header
#
# Note that you may have to ``--skip-header`` otherwise you might
# encounter some segfaults. There is currently a big mess of different
# versions of libraries in several levels of the MC file processings.
#
# The ``--ignore-hits`` will skip the hit information, so the converted file
# is much smaller (normally around 2-3 MBytes). Skip this option if you want
# to read the hit information too. The file will still be smaller than the
# ROOT file (about 1/3).
#
# Luckily, a handful people are preparing the HDF5 conversion, so in future
# you can download them directly, without thinking about which Jpp or Aanet
# version you need to open them.

#####################################################
# First Look at the Data
# ----------------------

filepath = "data/basic_analysis_sample.h5"

#####################################################
# We can have a quick look at the file with the ``ptdump`` command
# in the terminal::
#
#     ptdump filename.h5
#
# For further information, check out the documentation of the KM3NeT HDF5
# format definition: http://km3pipe.readthedocs.io/en/latest/hdf5.html
#

#####################################################
# The ``/event_info`` table contains general information about each event.
# The data is a simple 2D table and each event is represented by a single row.
#
# Let's have a look at the first few rows:
event_info = pd.read_hdf(filepath, '/event_info')
print(event_info.head(5))

#####################################################
# Next, we will read out the MC tracks which are stored under ``/mc_tracks``.

tracks = pd.read_hdf(filepath, '/mc_tracks')

#####################################################
# also read event info, for things like weights

info = pd.read_hdf(filepath, '/event_info')

#####################################################
# It has a similar structure, but now you can have multiple rows which belong
# to an event. The ``event_id`` column holds the ID of the corresponding event.

print(tracks.head(10))

#####################################################
# We now are accessing the first track for each event by grouping via
# ``event_id`` and calling the ``first()`` method of the
# ``Pandas.DataFrame`` object.

primaries = tracks.groupby('event_id').first()

#####################################################
# Here are the first 5 primaries:
print(primaries.head(5))

#####################################################
# Creating some Fancy Graphs
# --------------------------

#####################################################
#
plt.hist(primaries.energy, bins=100, log=True)
plt.xlabel('energy [GeV]')
plt.ylabel('number of events')
plt.title('Energy Distribution')

#####################################################
#
primaries.bjorkeny.hist(bins=100)
plt.xlabel('bjorken-y')
plt.ylabel('number of events')
plt.title('bjorken-y Distribution')

#####################################################
#
zeniths = kp.math.zenith(primaries.filter(regex='^dir_.?$'))
primaries['zenith'] = zeniths

plt.hist(np.cos(primaries.zenith), bins=21, histtype='step', linewidth=2)
plt.xlabel(r'cos($\theta$)')
plt.ylabel('number of events')
plt.title('Zenith Distribution')

#####################################################
#
# Starting positions of primaries
# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
plt.hist2d(primaries.pos_x, primaries.pos_y, bins=100, cmap='viridis')
plt.xlabel('x [m]')
plt.ylabel('y [m]')
plt.title('2D Plane')
plt.colorbar()

#####################################################
#
# If you have seaborn installed (`pip install seaborn`), you can easily create
# nice jointplots:
try:
    import seaborn as sns    # noqa
    km3pipe.style.use("km3pipe")    # reset matplotlib style
except:
    print("No seaborn found, skipping example.")
else:
    g = sns.jointplot('pos_x', 'pos_y', data=primaries, kind='hex')
    g.set_axis_labels("x [m]", "y[m]")
    plt.subplots_adjust(right=0.90)    # make room for the colorbar
    plt.title("2D Plane")
    plt.colorbar()
    plt.legend()

#####################################################
#
from mpl_toolkits.mplot3d import Axes3D    # noqa
fig = plt.figure()
ax = fig.add_subplot(111, projection='3d')
ax.scatter3D(primaries.pos_x, primaries.pos_y, primaries.pos_z, s=3)
ax.set_xlabel('x [m]', labelpad=10)
ax.set_ylabel('y [m]', labelpad=10)
ax.set_zlabel('z [m]', labelpad=10)
ax.set_title('3D Plane')

#####################################################
#
gandalfs = pd.read_hdf(filepath, '/reco/gandalf')
print(gandalfs.head(5))

#####################################################
#
gandalfs.columns

#####################################################
#
# plt.hist(gandalfs['lambda'], bins=50, log=True)
# plt.xlabel('lambda parameter')
# plt.ylabel('count')
# plt.title('Lambda Distribution of Reconstructed Events')

#####################################################
#
gandalfs['zenith'] = kp.math.zenith(gandalfs.filter(regex='^dir_.?$'))

plt.hist((gandalfs.zenith - primaries.zenith).dropna(), bins=100)
plt.xlabel(r'true zenith - reconstructed zenith [rad]')
plt.ylabel('count')
plt.title('Zenith Reconstruction Difference')

#####################################################
#
l = 0.2
lambda_cut = gandalfs['lambda'] < l
plt.hist((gandalfs.zenith - primaries.zenith)[lambda_cut].dropna(), bins=100)
plt.xlabel(r'true zenith - reconstructed zenith [rad]')
plt.ylabel('count')
plt.title('Zenith Reconstruction Difference for lambda < {}'.format(l))

#####################################################
# Combined zenith reco plot for different lambda cuts
# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

fig, ax = plt.subplots()
for l in [100, 5, 2, 1, 0.1]:
    l_cut = gandalfs['lambda'] < l
    ax.hist((primaries.zenith - gandalfs.zenith)[l_cut].dropna(),
            bins=100,
            label=r"$\lambda$ = {}".format(l),
            alpha=.7)
plt.xlabel(r'true zenith - reconstructed zenith [rad]')
plt.ylabel('count')
plt.legend()
plt.title('Zenith Reconstruction Difference for some Lambda Cuts')

#####################################################
# Fitting Angular resolutions
# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
#
# Let's fit some distributions: gaussian + lorentz (aka norm + cauchy)
#
# Fitting the gaussian to the whole range is a very bad fit, so
# we make a second gaussian fit only to +- 10 degree.
# Conversely, the Cauchy (lorentz) distribution is a near perfect fit
# (note that ``2 gamma = FWHM``).

from scipy.stats import cauchy, norm    # noqa

residuals = gandalfs.zenith - primaries.zenith
cut = (gandalfs['lambda'] < l) & (np.abs(residuals) < 2 * np.pi)
residuals = residuals[cut]
info[cut]

# convert rad -> deg
residuals = residuals * 180 / np.pi

pi = 180
# x axis for plotting
x = np.linspace(-pi, pi, 1000)

c_loc, c_gamma = cauchy.fit(residuals)
fwhm = 2 * c_gamma

g_mu_bad, g_sigma_bad = norm.fit(residuals)
g_mu, g_sigma = norm.fit(residuals[np.abs(residuals) < 10])

plt.hist(residuals, bins='auto', label='Histogram', density=True, alpha=.7)
plt.plot(
    x,
    cauchy(c_loc, c_gamma).pdf(x),
    label='Lorentz: FWHM $=${:.3f}'.format(fwhm),
    linewidth=2
)
plt.plot(
    x,
    norm(g_mu_bad, g_sigma_bad).pdf(x),
    label='Unrestricted Gauss: $\sigma =$ {:.3f}'.format(g_sigma_bad),
    linewidth=2
)
plt.plot(
    x,
    norm(g_mu, g_sigma).pdf(x),
    label='+- 10 deg Gauss: $\sigma =$ {:.3f}'.format(g_sigma),
    linewidth=2
)
plt.xlim(-pi / 4, pi / 4)
plt.xlabel('Zenith residuals / deg')
plt.legend()

####################################################################
# We can also look at the median resolution without doing any fits.
#
# In textbooks, this metric is also called Median Absolute Deviation.

resid_median = np.median(residuals)
residuals_shifted_by_median = residuals - resid_median
absolute_deviation = np.abs(residuals_shifted_by_median)
resid_mad = np.median(absolute_deviation)

plt.hist(np.abs(residuals), alpha=.7, bins='auto', label='Absolute residuals')
plt.axvline(resid_mad, label='MAD: {:.2f}'.format(resid_mad), linewidth=3)
plt.title("Average resolution: {:.3f} degree".format(resid_mad))
plt.legend()
plt.xlabel('Absolute zenith residuals / deg')



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# vim: ts=4 sw=4 et
"""
====================
Trigger Rate Monitor
====================

A (messy) script to monitor the trigger rates.

"""

from datetime import datetime
from collections import defaultdict, deque, OrderedDict
import sys
from io import BytesIO
import os
import shutil
import time
import threading

import matplotlib
# Force matplotlib to not use any Xwindows backend.
matplotlib.use('Agg')    # noqa
import matplotlib.pyplot as plt
import matplotlib.dates as md
import matplotlib.ticker as ticker

import km3pipe as kp
from km3pipe.config import Config
from km3pipe.io import CHPump
from km3pipe.io.daq import (DAQPreamble, DAQEvent)
import km3pipe.style
km3pipe.style.use('km3pipe')

log = kp.logger.get_logger("trigger_rate")

PLOTS_PATH = 'km3web/plots'

xfmt = md.DateFormatter('%Y-%m-%d %H:%M')
lock = threading.Lock()

general_style = dict(markersize=6, linestyle='None')
styles = {
    "Overall": dict(
        marker='D',
        markerfacecolor='None',
        markeredgecolor='tomato',
        markeredgewidth=1
    ),
    "3DMuon": dict(marker='X', markerfacecolor='dodgerblue'),
    "MXShower": dict(marker='v', markerfacecolor='orange'),
    "3DShower": dict(marker='o', markerfacecolor='greenyellow'),
}


def trigger_rate_sampling_period():
    try:
        return int(Config().get("Monitoring", "trigger_rate_sampling_period"))
    except (TypeError, ValueError):
        return 180


def is_3dshower(trigger_mask):
    return bool(trigger_mask & 1)


def is_mxshower(trigger_mask):
    return bool(trigger_mask & 2)


def is_3dmuon(trigger_mask):
    return bool(trigger_mask & 16)


class TriggerRate(kp.Module):
    def configure(self):
        self.run = True
        self.interval = self.get("interval") or trigger_rate_sampling_period()
        print("Update interval: {}s".format(self.interval))
        self.trigger_counts = defaultdict(int)
        self.trigger_rates = OrderedDict()
        for trigger in ["Overall", "3DMuon", "MXShower", "3DShower"]:
            self.trigger_rates[trigger] = deque(
                maxlen=int(60 * 24 / (self.interval / 60))
            )
        self.thread = threading.Thread(target=self.plot).start()

    def process(self, blob):
        if not str(blob['CHPrefix'].tag) == 'IO_EVT':
            return blob
        sys.stdout.write('.')
        sys.stdout.flush()

        data = blob['CHData']
        data_io = BytesIO(data)
        preamble = DAQPreamble(file_obj=data_io)    # noqa
        event = DAQEvent(file_obj=data_io)
        tm = event.trigger_mask
        with lock:
            self.trigger_counts["Overall"] += 1
            self.trigger_counts["3DShower"] += is_3dshower(tm)
            self.trigger_counts["MXShower"] += is_mxshower(tm)
            self.trigger_counts["3DMuon"] += is_3dmuon(tm)

        print(self.trigger_counts)

        return blob

    def plot(self):
        while self.run:
            time.sleep(self.interval)
            self.create_plot()

    def create_plot(self):
        print('\n' + self.__class__.__name__ + ": updating plot.")

        timestamp = datetime.utcnow()

        with lock:
            for trigger, n_events in self.trigger_counts.items():
                trigger_rate = n_events / self.interval
                self.trigger_rates[trigger].append((timestamp, trigger_rate))
            self.trigger_counts = defaultdict(int)

        fig, ax = plt.subplots(figsize=(16, 4))

        for trigger, rates in self.trigger_rates.items():
            timestamps, trigger_rates = zip(*rates)
            ax.plot(
                timestamps,
                trigger_rates,
                **styles[trigger],
                **general_style,
                label=trigger
            )
        ax.set_title(
            "Trigger Rates\n{0} UTC".format(datetime.utcnow().strftime("%c"))
        )
        ax.set_xlabel("time")
        ax.set_ylabel("trigger rate [Hz]")
        ax.xaxis.set_major_formatter(xfmt)
        ax.yaxis.set_major_locator(
            ticker.LogLocator(base=10.0, subs=(1.0, ), numticks=100)
        )
        ax.grid(True)
        ax.minorticks_on()
        plt.legend()

        fig.tight_layout()

        filename = os.path.join(PLOTS_PATH, 'trigger_rates_lin_test.png')
        filename_tmp = os.path.join(
            PLOTS_PATH, 'trigger_rates_lin_test_tmp.png'
        )
        plt.savefig(filename_tmp, dpi=120, bbox_inches="tight")
        shutil.move(filename_tmp, filename)

        try:
            ax.set_yscale('log')
        except ValueError:
            pass

        filename = os.path.join(PLOTS_PATH, 'trigger_rates_test.png')
        filename_tmp = os.path.join(PLOTS_PATH, 'trigger_rates_test_tmp.png')
        plt.savefig(filename_tmp, dpi=120, bbox_inches="tight")
        shutil.move(filename_tmp, filename)

        plt.close('all')
        print("Plot updated at '{}'.".format(filename))

    def finish(self):
        self.run = False
        if self.thread is not None:
            self.thread.stop()


pipe = kp.Pipeline()
pipe.attach(
    CHPump,
    host='127.0.0.1',
    port=5553,
    tags='IO_EVT',
    timeout=60 * 60 * 24 * 7,
    max_queue=200000
)
pipe.attach(TriggerRate, interval=60)
pipe.drain()



# Filename: runtable.py
"""
Prints the run table for a given detector ID.

Usage:
    runtable [options] DET_ID
    runtable (-h | --help)
    runtable --version

Options:
    -h --help           Show this screen.
    -c                  Compact view.
    -n RUNS             Number of runs.
    -r FROM_RUN-TO_RUN  Range of runs (example: 3100-3200).
    -s REGEX            Regular expression to filter the runsetup name/id.
    DET_ID              Detector ID (eg. D_ARCA001).

"""

import re
import sys
import km3pipe as kp

__author__ = "Tamas Gal"
__copyright__ = "Copyright 2018, Tamas Gal and the KM3NeT collaboration."
__credits__ = []
__license__ = "MIT"
__maintainer__ = "Tamas Gal"
__email__ = "tgal@km3net.de"
__status__ = "Development"

log = kp.logger.get_logger(__name__)


def runtable(det_id, n=5, run_range=None, compact=False, sep='\t', regex=None):
    """Print the run table of the last `n` runs for given detector"""
    db = kp.db.DBManager()
    df = db.run_table(det_id)

    if run_range is not None:
        try:
            from_run, to_run = [int(r) for r in run_range.split('-')]
        except ValueError:
            log.critical("Please specify a valid range (e.g. 3100-3200)!")
            raise SystemExit
        else:
            df = df[(df.RUN >= from_run) & (df.RUN <= to_run)]

    if regex is not None:
        try:
            re.compile(regex)
        except re.error:
            log.error("Invalid regex!")
            return

        df = df[df['RUNSETUPNAME'].str.contains(regex)
                | df['RUNSETUPID'].str.contains(regex)]

    if n is not None:
        df = df.tail(n)

    if compact:
        df = df[['RUN', 'DATETIME', 'RUNSETUPNAME']]

    df.to_csv(sys.stdout, sep=sep)


def main():
    from docopt import docopt
    args = docopt(__doc__, version=kp.version)

    try:
        n = int(args['-n'])
    except TypeError:
        n = None

    runtable(
        args['DET_ID'],
        n=n,
        run_range=args['-r'],
        regex=args['-s'],
        compact=args['-c']
    )



# Filename: test_widgets.py
"""
...

"""

from km3pipe.testing import TestCase, skip

from pipeinspector.widgets import BlobWidget

__author__ = "Tamas Gal"
__copyright__ = "Copyright 2016, Tamas Gal and the KM3NeT collaboration."
__credits__ = []
__license__ = "MIT"
__maintainer__ = "Tamas Gal"
__email__ = "tgal@km3net.de"
__status__ = "Development"


class TestBlobWidget(TestCase):
    def test_make_scale_labels(self):
        blobs = BlobWidget()
        blobs.width = 25
        scale_labels = blobs._make_scale_labels(0)
        self.assertEqual("0         10        20   ", scale_labels)
        scale_labels = blobs._make_scale_labels(2)
        self.assertEqual("0         10        20   ", scale_labels)
        scale_labels = blobs._make_scale_labels(10)
        self.assertEqual("0         10        20   ", scale_labels)
        scale_labels = blobs._make_scale_labels(11)
        self.assertEqual("         10        20    ", scale_labels)
        scale_labels = blobs._make_scale_labels(4589)
        self.assertEqual(" 4580      4590      ", scale_labels)

    @skip
    def test_make_ruler(self):
        blobs = BlobWidget()
        blobs.width = 25
        ruler = blobs._make_ruler(0)
        self.assertEqual("|    '    |    '    |    ", ruler)
        ruler = blobs._make_ruler(2)
        self.assertEqual("|    '    |    '    |    ", ruler)
        ruler = blobs._make_ruler(10)
        self.assertEqual("|    '    |    '    |    ", ruler)
        ruler = blobs._make_ruler(11)
        self.assertEqual("    '    |    '    |    '", ruler)
        ruler = blobs._make_ruler(12)
        self.assertEqual("   '    |    '    |    ' ", ruler)
        ruler = blobs._make_ruler(19)
        self.assertEqual(" |    '    |    '    |   ", ruler)
        ruler = blobs._make_ruler(20)
        self.assertEqual("|    '    |    '    |    ", ruler)
        ruler = blobs._make_ruler(23)
        self.assertEqual("  '    |    '    |    '  ", ruler)
        ruler = blobs._make_ruler(109)
        self.assertEqual(" |    '    |    '    |   ", ruler)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# vim: ts=4 sw=4 et
"""
=======================
K40 Calibration Summary
=======================

Combine k40calib results into a single CSV file.

.. code-block:: bash

    Usage:
        k40summary.py CALIB_FOLDER
        k40summary.py (-h | --help)

    Options:
        -h --help   Show this screen.

"""

from glob import glob
import os
import re
import pickle
import numpy as np


def write_header(fobj):
    """Add the header to the CSV file"""
    fobj.write("# K40 calibration results\n")
    fobj.write("det_id\trun_id\tdom_id")
    for param in ['t0', 'qe']:
        for i in range(31):
            fobj.write("\t{}_ch{}".format(param, i))


def main():
    from docopt import docopt
    args = docopt(__doc__)

    file_pattern = os.path.join(args['CALIB_FOLDER'], '*.p')
    files = glob(file_pattern)

    with open('k40calib_summary.csv', 'w') as csv_file:
        write_header(csv_file)

        for fn in files:
            det_id, run_id = [
                int(x) for x in re.search("_(\\d{8})" * 2, fn).groups()
            ]
            with open(fn, 'rb') as f:
                data = pickle.load(f)
            if not data:
                print("Empty dataset found for '{}'".format(fn))
            else:
                for dom_id in data.keys():
                    t0s = data[dom_id]['opt_t0s'].x
                    qes = data[dom_id]['opt_qes'].x
                    cols = np.concatenate([[det_id, run_id, dom_id], t0s, qes])
                    line = '\n' + '\t'.join(str(c) for c in cols)
                    csv_file.write(line)


if __name__ == '__main__':
    main()



#!/usr/bin/env python
# Filename: k40calib.py
# vim: ts=4 sw=4 et
"""
=========================
K40 Intra-DOM Calibration
=========================

The following script calculates the PMT time offsets using K40 coincidences


Usage:
    k40calib FILE DET_ID [-t TMAX -c CTMIN -r -o CALIB_FILE -s STREAM]
    k40calib (-h | --help)
    k40calib --version

Options:
    FILE            Input file (ROOT).
    DET_ID          Detector ID (e.g. 29).
    -r              Skip frames with with at least one PMT in HRV.
    -t TMAX         Coincidence time window [default: 10].
    -s STREAM       JDAQTimeslice stream (L1, L2, SN, ...) [default: ].
    -c CTMIN        Minimum cos(angle) between PMTs for L2 [default: -1.0].
    -o CALIB_FILE   Filename for the calibration output [default: k40_cal.p].
    -h --help       Show this screen.
"""
# Author: Jonas Reubelt <jreubelt@km3net.de> and Tamas Gal <tgal@km3net.de>
# License: MIT

from km3pipe import version
import km3pipe as kp
from km3modules import k40
from km3modules.common import StatusBar, MemoryObserver

__author__ = "Tamas Gal and Jonas Reubelt"
__copyright__ = "Copyright 2016, KM3NeT collaboration."
__credits__ = []
__license__ = "MIT"
__maintainer__ = "Jonas Reubelt and Tamas Gal"
__email__ = "tgal@km3net.de"
__status__ = "Development"


def k40calib(
    filename, tmax, ctmin, stream, filter_hrv, det_id, calib_filename
):
    pipe = kp.Pipeline()
    pipe.attach(kp.io.jpp.TimeslicePump, filename=filename, stream=stream)
    pipe.attach(StatusBar, every=5000)
    pipe.attach(MemoryObserver, every=10000)
    pipe.attach(
        k40.HRVFIFOTimesliceFilter, filter_hrv=filter_hrv, filename=filename
    )
    pipe.attach(k40.SummaryMedianPMTRateService, filename=filename)
    pipe.attach(k40.TwofoldCounter, tmax=tmax)
    pipe.attach(k40.K40BackgroundSubtractor, mode='offline')
    pipe.attach(
        k40.IntraDOMCalibrator,
        ctmin=ctmin,
        mode='offline',
        det_id=det_id,
        calib_filename=calib_filename
    )
    pipe.drain()


def main():
    from docopt import docopt
    args = docopt(__doc__, version=version)

    k40calib(
        args['FILE'], int(args['-t']), float(args['-c']), args['-s'],
        args['-r'], int(args['DET_ID']), args['-o']
    )
